/*
 *  NXPostScript - NXPostScript.h
 *
 *  $RCSfile: NXPostScript.h,v $
 *  $Revision: 1.5 $
 *  $Date: 1993/12/10 04:14:22 $
 *
 */
#ifndef _PSLIB_H_
#define _PSLIB_H_

#include <stdio.h>

#include <Xm/Xm.h>

/***********************************************************/

/*
 * These are the sheet format values that can be specified.
 */
#define XnPS_PORTRAIT		200	/* Portrait sheet format. */
#define	XnPS_LANDSCAPE		210	/* Landscape sheet format. */
#define	XnPS_TWO_ONE		220	/* Two pages on one sheet of paper. */
#define	XnPS_FOUR_ONE		230	/* Four pages on one sheet of paper. */

/*
 * These are error values that may be returned by the library functions that return values.
 */
#define	XnPS_READ_ERROR		500	/* There was an error reading from the input file. */
#define	XnPS_WRITE_ERROR	501	/* There was an error reading from the output file. */
#define	XnPS_ILLEGAL_SRC	502	/* The src file was not valid. */
#define	XnPS_ILLEGAL_DEST	503	/* The dest file was not valid. */
#define	XnPS_ILLEGAL_PROPS	504	/* The properties structure was not valid. */

/***********************************************************/

/*
 * This the the structure that contains the properties that the generated PostScript
 * file will follow.  A structure containing default values can be created with the
 * XnCreatePSProps library call. Each element is marked as follows:
 *
 *	'-' Should NOT be modified by the application.
 *	'M' Must be modified by the application.
 *	'C' Can be modified to taylor output style.
 *	'R' Rarely modified by the application.
 *
 */
typedef struct _XnPSPropsStruct XnPSPropsStruct;
struct _XnPSPropsStruct {

       /*
	* - 	xScale
	* -	yScale
	* M	sheetFormat		- The sheet format (defined above).
	* -	needHeaderArea
	*/
	float xScale, yScale;
	int sheetFormat;
	Boolean needHeaderArea;

       /*
	* C	tabWidth		- The number of spaces to expand tabs.
	* R	maxPageNumber		- The maximum page number (for spacing).
	* C	maxLineNumber		- The number of lines in the file (for spacing).
	* C	leadingSpaces		- The 'wrap' leading spaces.
	* -	charWidths
	* R	borderLineWidth		- The point size of the border lines.
	* R	headerLineWidth		- The point size of the header lines.
	* R	headerBackground	- The shading of the header background.
	* M	printRangeFrom		- The page number to start printing from (see printPageRange).
	* M	printRangeTo		- The page number to stop printing at (see printPageRange).
	*/
	int tabWidth, maxPageNumber, maxLineNumber, printRangeFrom, printRangeTo;
	char *leadingSpaces;
	float *charWidths, borderLineWidth, headerLineWidth, headerBackground;

       /*
	* Offsets:
	*
	* R	topOffsetPts		- Points to offset from top of page.
	* R	bottomOffsetPts		- Points to offset from bottom of page.
	* R	bodyRightOffsetPts	- Points to offset text from right of page.
	* R	bodyLeftOffsetPts	- Points to offset text from left of page.
	*
	* R	headerLeftOffsetPts	- Points to offset header from left of page. 
	* R	headerLineTopOffsetPts	- Points to offset header from top of page.
	* R	headerLineBottomOffsetPts - Points to offset header from bottom of page.
	*
	* R	borderTopOffsetPts	- Points to offset top border.
	* R	numberingLineOffset	- Points to offset line numbers from text.
	*/
	int topOffsetPts, bottomOffsetPts, bodyRightOffsetPts, bodyLeftOffsetPts;
	int headerLeftOffsetPts, headerLineTopOffsetPts, headerLineBottomOffsetPts;
	int borderTopOffsetPts, numberingLineOffset;

       /*
	* Page Dimensions:
	*
	* -	virtualPageWidthPts
	* -	virtualPageHeightPts
	* C	physicalPageWidthPts	- Output sheet width in points (1" = 72pts).
	* C	physicalPageHeightPts	- Output sheet height in points.
	*/
	int virtualPageWidthPts, virtualPageHeightPts;
	int physicalPageWidthPts, physicalPageHeightPts;

       /*
	* Margins:
	*
	* C	topMarginPts		- Top margin in points (1" = 72pts).
	* C	leftMarginPts		- Left margin in points.
	* C	rightMarginPts		- Right margin in points.
	* C	bottomMarginPts		- Bottom margin in points.
	* C	innerMarginLRPts	- Inner margin left/right in points (two / four format).
	* C	innerMarginTBPts	- Inner margin top/bottom in points (two / four format).
	*/
	int topMarginPts, bottomMarginPts, leftMarginPts, rightMarginPts;
	int innerMarginLRPts, innerMarginTBPts;

       /*
	* Font Attributes:
	*
	* C	titleFontSize		- The font size for the document title.
	* C	headerFontSize		- The font size for the header.
	* C	bodyFontSize		- The font size for the main text.
	* C	bodyFont		- The font for the main text.
	* C	boldFont		- The font for bold text.
	* C	titleFont		- The font for the document title.
	* C	italicFont		- The font for italic text.
	* C	headerFont		- The font for the header.
	*/
	int titleFontSize, headerFontSize, bodyFontSizePts;
	char *bodyFont, *boldFont, *titleFont, *italicFont, *headerFont;

       /*
	* M	time			- The current time.
	* M	user			- Who the PostScript file is for.
	* M	title			- The title of the PostScript file.
	* M	userID			- The login of the user.
	* M	creator			- The application creating the file.
	* M	srcFilename		- The name of the source file.
	* M	creationDate		- The creation date.
	*/
	char *time, *user, *title, *userID, *creator, *srcFilename, *creationDate;

       /*
	* Boolean Attributes:
	*
	* C	allPages		- Print options on all pages?
	* C	printTime		- Print the time in the header?
	* C	printTitle		- Print the title in the header?
	* C	printUserID		- Print the user id in the header?
	* C	printBorder		- Print a border around the page?
	* C	characterWrap		- Perform character wrapping?
	* C	processEscape		- Process vt100 escape characters?
	* C	printTotalPages		- Print the total number of pages in header?
	* C	printLineNumbers	- Print the line numbers?
	* C	printPageNumbers	- Print page numbers in the header?
	* C	printPageRange		- Print a certain range of pages?
	* C	printCredits		- Print the credits on each page.
        */
	Boolean allPages, printTime, printTitle, printUserID, printBorder;
	Boolean characterWrap, processEscape, printTotalPages;
	Boolean printLineNumbers, printPageNumbers, printPageRange;
	Boolean printCredits;

	};

/***********************************************************/

/*
 * This structure is contains the PostScript output statistics
 * and is filled by the XnGeneratePostScript library function.
 *
 *	imageCnt		- The number of PostScript images converted.
 *	showPageCnt		- The number of 'showpage' calls in the PostScript.
 *	linesWrapped		- The number of lines that were wrapped.
 *	charsProcessed		- The number of characters processed.
 *	linesProcessed		- The number of lines processed.
 *	physicalPagesProcessed	- The number of pages in the PostScript file.
 *	virtualPagesProcessed	- The number of pages processed.
 *	actualBodyTextWidth	- Width of page that was used for text.
 *	longestLineWidth	- The width of the longest line processed.
 *
 */
typedef struct _XnPSOutputPropsStruct XnPSOutputPropsStruct;
struct _XnPSOutputPropsStruct {
	int		imageCnt;
	int		showPageCnt;
	int		linesWrapped;
	int		charsProcessed;
	int		linesProcessed;
	int		physicalPagesProcessed;
	int		virtualPagesProcessed;
	int		actualBodyTextWidth;

	float		longestLineWidth;
	};

/***********************************************************/

/*
 * Given a fontname and a style, returns the corresponding PostScript
 * fontname (e.g. "times" "bold" == "Times-Bold"). The styles allowed are:
 *
 *	"NORMAL", "BOLD", "ITALIC", "BOLDITALIC"
 *
 * The fonts allowed are:
 *
 *	"TIMES", "HELVETICA", "COURIER", "AVANTGARDE", "PALATINO", "NEWCENTURYSCHLBK",
 *	"LUBALINGRAPH", "HELVETICA-NARROW", "SOUVENIR", "BOOKMAN", "ZAPFCHANCERY"
 *
 * Returns NULL if there is no corresponding PostScript font.
 */
extern char		*XnCreatePSFontname		( char *fontname,
							  char *fontstyle );


/*
 * Converts an ASCII input file into a PostScript output file.
 * The file descriptors 'src' and 'dest' should be two files opened
 * for read and write only (respectively).  The function starts reading
 * from the 'src' where the file descriptor specifies and outputs the generated
 * PostScript to 'dest'.  There are no seeks  in the function, strictly reads
 * and writes. The 'p' structure contains the PostScript file properties such
 * as what sheet style to use, font style/size, etc...  The output statistics
 * are returned in the 'p' structure (which must have space allocated already).
 * The function returns 0 if no errors occur, or an error code defined above
 * otherwise.
 */
extern int		XnGeneratePostScript		( FILE *src,
							  FILE *dest,
							  XnPSPropsStruct *p,
						 	  XnPSOutputPropsStruct *o );


/*
 * Tries to determine the number of pages in the PostScript file specified
 * by 'fileName'.  Returns the number of pages in the file or -1 if it cannot
 * be determined.
 */
extern int		XnPagesInPostScriptFile		( char *fileName );


/*
 * Given a PostScript font name (e.g. "Helvetica-Bold") it returns a pointer
 * to an array of character width values (e.g. The width of character 'A' is
 * in the 65th array position).  DO NOT FREE THE POINTER!  The pointer
 * returned is static and should not be modified.
 */
extern float		*XnPSFontCharWidths		( char *fontname );


/*
 * Tries to determine if the specified file is a PostScript file or not.
 * Returns True if the file is PostScript and False otherwise.
 */
extern Boolean		XnIsPostScriptFile		( char *fileName );


/*
 * Creates and returns a default property structure based on the
 * specified sheet format (defined above). This pointer should be passed
 * as an argument to XnGeneratePostScript.
 */
extern XnPSPropsStruct	*XnCreatePSProps		( int sheetFormat );


/*
 * Modifies the structure elements x, y scale and virtual page width,height
 * based on changes in the physical page width,height settings.  Should be
 * called if an application modifies any elements in the default properties
 * structure.
 */
extern void		XnPSSetVirtualPageDimensions	( XnPSPropsStruct *p );

/*
 * Converts a pixmap to PostScript and outputs the code to a destination
 * file.  The PostScript code that is output is designed to be imbedded
 * within a NXPostScript driver "control" text file.  The text file would
 * ordinarily contain ASCII text with control characters (defined by the driver).
 * Three control characters notify the driver to imbed sections of PostScript code
 * without manipulation.  This is the intent of this function - to produce
 * the imbeded image code.  The image code output is not independent.  It 
 * is not followed by a 'showpage' or any other page descriptors.  It is only
 * the PostScript to reproduce the image.
 */
extern int		NXPixmapToPostScript		( Display *display,
							  FILE *destFile,
							  Pixmap bitmap,
							  unsigned int bits_per_sample,
							  unsigned int pixmapWidth,
							  unsigned int pixmapHeight,
							  unsigned int boundingBoxWidth,
							  unsigned int boundingBoxHeight );

#endif
