/*
 *  NXPostScript - nxpsdriver.c 
 *
 *	Version 2.0
 *      by Lance Lovette
 *	Code and Display Copyright (C) 1993 by Lance Lovette
 *	All Rights Reserved
 *
 *	Lance Lovette (lovette@iftech.com)
 *	North Carolina State University
 *	Engineering Computer Operations
 *	Department of Computer Science
 *	Raleigh, NC 27607
 *
 *  Permission to use, copy, modify and distribute this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both the
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the names of Lance Lovette and North Carolina
 *  State University not be used in advertising or publicity pertaining to
 *  distribution of the software without specific, written prior permission.
 *  Lance Lovette and North Carolina State University make no
 *  representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 *  Lance Lovette AND North Carolina State University DISCLAIMS ALL
 *  WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Lance Lovette OR
 *  North Carolina State University BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 *  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 *  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 *  OF THIS SOFTWARE.
 *
 *  $RCSfile: nxpsdriver.c,v $
 *  $Revision: 2.2 $
 *  $Date: 1994/01/02 20:01:03 $
 */
static char rcsid[] = "$Id: nxpsdriver.c,v 2.2 1994/01/02 20:01:03 ltlovett Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <errno.h>

#include <Xm/Xm.h>

#include "NXUtils.h"
#include "NXPostScript.h"
#include "nxpsutils.h"

#define	EMAIL			"lovette@iftech.com"

#define	YEAR			1993		/* The current year. */
#define	VERSION_STR		"3.1"		/* The current version of XnGeneratePostScript. */
#define	MAX_LINE_LEN		512		/* Maximum length of a line in the input file. */

#define	TEXT_BOLD		600		/* When the currentEscapeType is Bold mode. */
#define	TEXT_ITALIC		601		/* When the currentEscapeType is Italic mode. */
#define	TEXT_UNDERLINE		602		/* When the currentEscapeType is Underline mode. */
#define	TEXT_NORMAL		603		/* When the currentEscapeType is Normal mode. */
#define	IMAGE_CENTER		700		/* When the currentEscapeType is Center mode. */
#define	IMAGE_LEFT		701		/* When the currentEscapeType is Left mode. */
#define	IMAGE_RIGHT		702		/* When the currentEscapeType is Right mode. */

#define	TAB			'\t'		/* The tab character in an input file. */
#define	ESCAPE			27		/* The escape character in an input file. */
#define	FORMFEED		'\f'		/* A formfeed character in an input file. */
#define	NEWLINE			'\n'		/* A newline character in an input file. */

#define	IMAGE_CENTER_ESCAPE	'c'		/* Escape character to start image center. */
#define	IMAGE_LEFT_ESCAPE	'l'		/* Escape character to start image left. */
#define	IMAGE_RIGHT_ESCAPE	'r'		/* Escape character to start image right. */
#define	TEXT_UNDERLINE_ESCAPE	'4'		/* Escape character to start underlining text. */
#define	TEXT_BOLD_ESCAPE	'1'		/* Escape character to start bolding text. */
#define	TEXT_ITALIC_ESCAPE	'7'		/* Escape character to start italicizing text. */
#define	TEXT_NORMAL_ESCAPE	'm'		/* Escape character to restore to normal mode. */

/*
 * The following are the names of procedures that are output
 * in the PostScript file.  They are kept short for reasons
 * of PostScript interpretation efficiency.  Their definitions
 * appear in the OutputHeaderPS function.
 */
#define	SHOW_LINE_PROC		"LJ"		/* Show a line of text. */
#define	DRAW_RECTANGLE_PROC	"DR"		/* Draw a rectangle. */
#define	DRAW_LINE_PROC		"DL"		/* Draw a line. */
#define	DRAW_HEADER_BOX_PROC	"DHB"		/* Draw the header box. */
#define	DRAW_TITLE_PROC		"DT"		/* Draw the title. */
#define	DRAW_USERID_PROC	"DUID"		/* Draw the user id. */
#define	DRAW_TIME_PAGE_PROC	"DTP"		/* Draw the time and page number. */
#define	SET_FONT_PROC		"SF"		/* Set the current font. */
#define	SHOW_LINE_NUM_PROC	"SLN"		/* Show the current line number. */

/*
 * The following definitions are used in the code to shorten the
 * statements that use them, and to make necessary changes to their
 * common definitions simple.  All values are in points. Descriptions follow:
 *
 *  NUMBERING_LINE_OFFSET	-An offset around the text line numbers for spacing.
 *  LEADING_SPACES		-The text that is used as leading spaces when a line is wrapped.
 *  TOP_THRESHOLD		-The y coordinate of the very top of the printed page.
 *  BOTTOM_THRESHOLD		-The y coordinate of the very bottom of the printed page.
 *  BODY_LEFT_THRESHOLD		-The x coordinate where the text in the main body will begin.
 *  HEADER_LEFT_THRESHOLD	-The x coordinate where the text in the header will begin.
 *  RIGHT_THRESHOLD		-The x coordinate where the text in the main body will end.
 *  BODY_LINE_HEIGHT		-The height of each line in the main body of text. 
 *  HEADER_THICKNESS		-The thickness of the header at the top of each page.
 *  HEADER_OFFSET		-The thickness of the header plus some spacing at the bottom of the header
 *  USER_ID_WIDTH_PTS		-The width (in points) of the header user id section. (144 pts)
 *  TITLE_WIDTH_PTS		-The width (in points) of the header title width section. (210 pts)
 *  TIME_PAGE_WIDTH_PTS		-The width (in points) of the header time & page number width section. (156 pts)
 *  PTS_PER_INCH		-Number of points/inch in PostScript (standard).
 */
#define	NUMBERING_LINE_OFFSET	(d_psProps->numberingLineOffset)
#define	LEADING_SPACES		d_psProps->leadingSpaces
#define	TOP_THRESHOLD		((d_psProps->bottomMarginPts + d_psProps->virtualPageHeightPts) - d_psProps->topOffsetPts)
#define	BOTTOM_THRESHOLD	(d_psProps->bottomMarginPts + d_psProps->bottomOffsetPts)
#define	BODY_LEFT_THRESHOLD	(d_psProps->leftMarginPts + d_psProps->bodyLeftOffsetPts)
#define	HEADER_LEFT_THRESHOLD	(d_psProps->leftMarginPts + d_psProps->headerLeftOffsetPts)
#define	RIGHT_THRESHOLD		((d_psProps->leftMarginPts + d_psProps->virtualPageWidthPts) - d_psProps->bodyRightOffsetPts)
#define	BODY_LINE_HEIGHT	(d_psProps->bodyFontSizePts)
#define	HEADER_THICKNESS	(d_psProps->borderTopOffsetPts + d_psProps->titleFontSize + d_psProps->headerLineTopOffsetPts)
#define	HEADER_OFFSET		(HEADER_THICKNESS + d_psProps->headerLineWidth + d_psProps->headerLineBottomOffsetPts)
#define	USER_ID_WIDTH_PTS	( (int)((float)d_psProps->virtualPageWidthPts * 0.2667) )
#define	TITLE_WIDTH_PTS		( (int)((float)d_psProps->virtualPageWidthPts * 0.3889) )
#define	TIME_PAGE_WIDTH_PTS	( (int)((float)d_psProps->virtualPageWidthPts * 0.2889) )
#define	PTS_PER_INCH		72

/*
 * PPIF - Points-per-inch as a float.
 * PSCHARWIDTH - The width of a character in points.
 * PROCESSING_IMAGE - True if the text being processed is being inlined (not manipulated).
 */
#define	PPIF( x )		( (float)((float)(x) / (float)(PTS_PER_INCH)) )
#define	PSCHARWIDTH( c )	( (float)((float)d_psProps->charWidths[ (int)(c) ] * (float)d_psProps->bodyFontSizePts) )
#define	PROCESSING_IMAGE(x)	( (x==IMAGE_CENTER) || (x==IMAGE_LEFT) || (x==IMAGE_RIGHT) )

/*
 * The following definitions allow the program to output a selected range of
 * pages in PostScript.  Descriptions follow:
 *
 *  IN_RANGE	-Evaluates to True if the user wants to print a range and the current page number is in that range.
 *  PAGE_RANGE	-Evaluates to True if the user does not want to print a range OR they want to and PAGE_RANGE is True.
 *  if_fprintf	-Performs a fprintf if PAGE_RANGE is true.
 *  if_putc	-Performs a putc if PAGE_RANGE is true.
 */
#define	IN_RANGE		( (d_psProps->printPageRange == True) && (d_physicalPageNum >= d_psProps->printRangeFrom) && (d_physicalPageNum <= d_psProps->printRangeTo) )
#define	PAGE_RANGE		( (d_psProps->printPageRange == False ) || IN_RANGE )
#define	if_fprintf		if( PAGE_RANGE ) fprintf
#define	if_putc(a,b)			if( PAGE_RANGE ) putc(a,b)

/*
 * Performs Integer and Float divisions and casts them to specific types.
 *
 *  I_HALF	-Divides a number by 2 and returns the integer result.
 *  I_QUART	-Divides a number by 4 and returns the integer result.
 *  I_DIV	-Divides x by y and returns the integer result.
 *  F_DIV	-Divides x by y and returns the floating point result.
 */
#define	I_HALF(x)		( (int)   ((int)x / (int)2) )
#define	I_QUART(x)		( (int)   ((int)x / (int)4) )
#define	I_DIV(x, y)		( (int)   ((float)(x) / (float)(y)) )
#define	F_DIV(x, y)		( (float) ((float)(x) / (float)(y)) )

/***********************************************************/

/*
 * If a file error occurs as defined by ferror() while reading or writing
 * the PostScript file, then a value of XnPS_READ_ERROR or XnPS_WRITE_ERROR
 * is returned by the function.  If g_errorType is not zero (0) when caught in
 * XnGeneratePostScript then the value is returned and the process terminates.
 */
int	g_errorType;

/*
 * The following variables contain the current statistics on the PostScript
 * output files.  The values of these variables are returned to the user
 * upon successful completion of XnGeneratePostScript.  They are described below:
 */
int	g_startingPageNum;		/* What is the first (physical) page number being output. */
int	g_showPageCnt;			/* The number of showpage's that are output. */
int	g_imageCnt;

/***********************************************************/

/*
 *  Proc:	XnPSSetVirtualPageDimensions
 *  Callers:	XnCreatePSProps
 *  Purpose:	Based on the specified sheet format, page width, and page height,
 *		calculate the appropriate scale factors and virtual page dimensions
 *		and store the values back into the structure.
 *  Params:	ps - The properties structure containing the values.
 *  Globals:	None
 *  Returns:	None
 */
void XnPSSetVirtualPageDimensions( XnPSPropsStruct *ps )
{
	int	margins;
	int	lrMargin;
	int	tbMargin;

	if( !ps ) return;

	lrMargin = ps->leftMarginPts + ps->rightMarginPts;
	tbMargin = ps->topMarginPts + ps->bottomMarginPts;

	switch( ps->sheetFormat )
	{
	case XnPS_PORTRAIT:
		ps->xScale			= 1.0;
		ps->yScale			= 1.0;
		ps->virtualPageWidthPts		= ps->physicalPageWidthPts - lrMargin;
		ps->virtualPageHeightPts	= ps->physicalPageHeightPts - tbMargin;
		break;

	case XnPS_LANDSCAPE:
		ps->xScale			= 1.0;
		ps->yScale			= 1.0;
		ps->virtualPageWidthPts		= ps->physicalPageHeightPts - lrMargin;
		ps->virtualPageHeightPts	= ps->physicalPageWidthPts - tbMargin;
		break;

	/*
	 * How the virtual page width / height is computed...
	 * The virtualPageWidth width is set manually.  We want to be able to fit 2 sheets
	 * of the virtual page plus margins (left, right, inner) in the space of the HEIGHT
	 * of the physical sheet.  There must be a scale associated with the virtual sheet to
	 * make it fit.  The formula (where x is the scale) is:
	 * leftMargin + (virtualPageWidth)x + innerMargin + (virtualPageWidth)x + rightMargin = physicalPageHeight
	 * Solve this for x.  Now that we have an xScale we want the yScale to be the same
	 * so the aspect ratio of width to height is the same.  We want the virtual height
	 * to be able to fit on the WIDTH of the physical sheet minus margins.  The formula is:
	 * bottomMargin + (virtualPageHeight)x + topMargin = physicalPageWidth
	 * Solve this for virtualPageHeight.
	 */
	case XnPS_TWO_ONE:
		margins = lrMargin + ps->innerMarginLRPts;

		ps->virtualPageWidthPts		= ps->physicalPageWidthPts;
		ps->xScale			= F_DIV( (ps->physicalPageHeightPts - margins), (2 * ps->virtualPageWidthPts) );

		margins = tbMargin;

		ps->yScale			= ps->xScale;
		ps->virtualPageHeightPts	= I_DIV( (ps->physicalPageWidthPts - margins), ps->xScale );
		break;

	/*
	 * How the virtual page width / height is computed...
	 * The virtualPageWidth width is set manually.  We want to be able to fit 2 sheets
	 * of the virtual page plus margins (left, right, inner) in the space of the WIDTH
	 * of the physical sheet.  There must be a scale associated with the virtual sheet to
	 * make it fit.  The formula (where x is the scale) is:
	 * leftMargin + (virtualPageWidth)x + innerMargin + (virtualPageWidth)x + rightMargin = physicalPageWidth
	 * Solve this for x.  Now that we have an xScale we want the yScale to be the same
	 * so the aspect ratio of width to height is the same.  We want to fit 2 sheets with the virtual height
	 * to be able to fit on the HEIGHT of the physical sheet minus margins.  The formula is:
	 * bottomMargin + (virtualPageHeight)x + innerMargin + (virtualPageHeight)x + topMargin = physicalPageHeight
	 * Solve this for virtualPageHeight.
	 */
	case XnPS_FOUR_ONE:
		margins = lrMargin + ps->innerMarginLRPts;

		ps->virtualPageWidthPts		= ps->physicalPageWidthPts;
		ps->xScale			= F_DIV( (ps->physicalPageWidthPts - margins), (2 * ps->virtualPageWidthPts) );

		margins = tbMargin + ps->innerMarginTBPts;

		ps->yScale			= ps->xScale;
		ps->virtualPageHeightPts	= I_DIV( (ps->physicalPageHeightPts - margins), (2 * ps->xScale) );
		break;
	}
}

/***********************************************************/

/*
 *  Proc:	XnCreatePSProps
 *  Callers:	<Library>
 *  Purpose:	Create and return a properties structure with default values based
 *		on the specified sheet format.
 *  Params:	sheetFormat - The sheet format to base the defaults on.
 *  Globals:	g_DefaultWidths, g_CourierWidths
 *  Returns:	Pointer to a structure containing default values.
 */
XnPSPropsStruct *XnCreatePSProps( int sheetFormat )
{
	XnPSPropsStruct	*ps;

	ps=XtNew( XnPSPropsStruct );

	ps->sheetFormat			= sheetFormat;
	ps->tabWidth			= 8;
	ps->maxPageNumber		= 8;
	ps->maxLineNumber		= 999;
	ps->headerBackground		= 0.97;
	ps->numberingLineOffset		= 5;
	ps->creator			= NULL;
	ps->title			= "NXpr";
	ps->leadingSpaces		= ".....";
	ps->user			= NULL;
	ps->userID			= NULL;
	ps->time			= NULL;
	ps->creationDate		= NULL;
	ps->srcFilename			= NULL;
	ps->printRangeFrom		= 0;
	ps->printRangeTo		= 0;

	ps->processEscape		= True;
	ps->printTotalPages		= False;
	ps->printLineNumbers		= False;
	ps->characterWrap		= True;
	ps->printPageNumbers		= True;
	ps->printTime			= True;
	ps->printTitle			= True;
	ps->printUserID			= True;
	ps->allPages			= True;
	ps->printPageRange		= False;
	ps->printCredits		= True;

	ps->bodyFont			= "Courier";
	ps->headerFont			= "Courier";
	ps->titleFont			= "Courier-Bold";
	ps->boldFont			= "Courier-Bold";
	ps->italicFont			= "Courier-Oblique";
	ps->bodyFontSizePts		= 10;
	ps->headerFontSize		= 10;

	ps->physicalPageWidthPts	= 612;
	ps->physicalPageHeightPts	= 792;

	ps->topMarginPts		= 36;
	ps->bottomMarginPts		= 36;
	ps->leftMarginPts		= 36;
	ps->rightMarginPts		= 36;
	ps->innerMarginLRPts		= 18;
	ps->innerMarginTBPts		= 18;

	ps->topOffsetPts		= 2;
	ps->bottomOffsetPts		= 3;
	ps->bodyLeftOffsetPts		= 3;
	ps->bodyRightOffsetPts		= 3;

	ps->borderTopOffsetPts		= 0;
	ps->headerLineWidth		= 1.0;
	ps->headerLineTopOffsetPts	= 3;
	ps->headerLineBottomOffsetPts	= 3;
	ps->headerLeftOffsetPts		= 3;

	switch( sheetFormat )
	{
	case XnPS_PORTRAIT:
		ps->titleFontSize		= ps->headerFontSize + 2;
		ps->borderLineWidth		= 0.5;
		ps->printBorder			= False;
		break;

	case XnPS_LANDSCAPE:
		ps->titleFontSize		= ps->headerFontSize + 2;
		ps->borderLineWidth		= 0.5;
		ps->printBorder			= False;
		break;

	case XnPS_TWO_ONE:
		ps->titleFontSize		= ps->headerFontSize + 4;
		ps->printBorder			= True;
		ps->borderLineWidth		= 1.0;
		break;

	case XnPS_FOUR_ONE:
		ps->titleFontSize		= ps->headerFontSize + 4;
		ps->printBorder			= True;
		ps->borderLineWidth		= 1.0;
		break;
	}

	XnPSSetVirtualPageDimensions( ps );

	/* Set when GeneratePostScript is called. */
	ps->needHeaderArea = False;

	if( ps->characterWrap == True )
		ps->charWidths = g_CourierWidths;
	else
		ps->charWidths = g_DefaultWidths;

	return( ps );
}

/***********************************************************/

/*
 *  Proc:	BodyTextWidthPts
 *  Callers:	XnGeneratePostScript
 *  Purpose:	Determine the width in points of a character string using
 *		the current font and font size.
 *  Params:	string - The character string.
 *		d_psProps - The PostScript output properties (font/size).
 *  Globals:	None
 *  Returns:	The width of the string in points.
 */
int BodyTextWidthPts( char *string, XnPSPropsStruct *d_psProps )
{
	int	i	=0;
	float	width	=0.0;

	if( !string ) return( -1 );

	for( i=0, width=0.0; string[i]; i++ )
		width += PSCHARWIDTH( string[i] );

	if( width < 0.0 ) width = (float)strlen( string ) * (float)d_psProps->bodyFontSizePts * 0.7;

	return( (int)width );
}

/***********************************************************/

/*
 *  Proc:	OutputHeaderPS
 *  Callers:	XnGeneratePostScript
 *  Purpose:	Output the header information at the beginning of the
 *		PostScript output file.
 *  Params:	destFile - The output file.
 *		d_psProps - The output PostScript properties.
 *  Globals:	g_errorType
 *  Returns:	None
 */
void OutputHeaderPS( FILE *destFile, XnPSPropsStruct *d_psProps )
{
	/* Strip the newline off the creation date. */
	if( d_psProps->creationDate && (d_psProps->creationDate[ strlen(d_psProps->creationDate) - 1 ] == '\n') )
		d_psProps->creationDate[ strlen(d_psProps->creationDate) - 1 ] = '\0';

	if( d_psProps->title && (d_psProps->title[ strlen(d_psProps->title) - 1 ] == '\n') )
		d_psProps->title[ strlen(d_psProps->title) - 1 ] = '\0';

	fprintf( destFile, "%%!PS-Adobe-2.0\n" );
	fprintf( destFile, "%%%%BoundingBox: 0 0 612 792\n" );
	fprintf( destFile, "%%%%Creator: %s\n", d_psProps->creator );
	fprintf( destFile, "%%%%Title: %s\n", d_psProps->title );
	fprintf( destFile, "%%%%For: %s\n", d_psProps->user );
	fprintf( destFile, "%%%%CreationDate: %s\n", d_psProps->creationDate );
	fprintf( destFile, "%%%%DocumentFonts: %s %s %s %s %s\n",
					d_psProps->bodyFont, d_psProps->headerFont, d_psProps->titleFont,
					d_psProps->boldFont, d_psProps->italicFont );
	fprintf( destFile, "%%%%Pages: (atend)\n");

	fprintf( destFile, "\n");
	fprintf( destFile, "%% NXPostScript Generator %s\n",  VERSION_STR );
	fprintf( destFile, "%% Copyright (C) %d by Lance Lovette\n", YEAR );
	fprintf( destFile, "%% Engineering Computer Operations\n" );
	fprintf( destFile, "%% North Carolina State University\n" );
	fprintf( destFile, "%% %s\n", EMAIL );

	fprintf( destFile, "%% Source file: %s\n", d_psProps->srcFilename);
	fprintf( destFile, "%% Sheet Format: %d\n", d_psProps->sheetFormat);
	fprintf( destFile, "%% Virtual Page Width:  %.2f\n", PPIF( d_psProps->virtualPageWidthPts) );
	fprintf( destFile, "%% Virtual Page Height: %.2f\n", PPIF( d_psProps->virtualPageHeightPts) );
	fprintf( destFile, "%% Margins: %.2f %.2f %.2f %.2f\n",
					PPIF(d_psProps->topMarginPts),
					PPIF(d_psProps->bottomMarginPts),
					PPIF(d_psProps->leftMarginPts),
					PPIF(d_psProps->rightMarginPts) );
	fprintf(destFile, "%%%%EndComments\n" );

	fprintf( destFile, "\n" );
	fprintf( destFile, "%% Define some procedures.\n" );

	/*
	 * Arguments are given in the order they are pushed on the stack in PostScript. 
	 *
	 * LJ (Left Justify String) arguments: text, x, y
	 * CS (Center String) arguments:  text, x, y
	 * RJ (Right Justify String) arguments:  text, x, y
	 */
	fprintf( destFile, "/LJ    { moveto show } bind def \n");
	fprintf( destFile, "/RJ    { moveto dup stringwidth pop neg 0 rmoveto show } bind def \n");
	fprintf( destFile, "/CS    { moveto dup stringwidth 2 div neg exch 2 div neg exch rmoveto show } bind def \n");

	/*
	 * SF (Set Font) font size, font name
	 * DL (Draw Line) x, y start  x, y end
	 * DR (Draw Rectangle) arguments: width, height, LLeft X, LLeftY
	 * RL (Roll)
	 * CP (Current Point)
	 * ULB (Underline begin)
	 * ULE (Underline end)
	 * BLDB (Bold text begin)
	 * BLDE (Bold text end)
	 * ITLB (Italic Begin)
	 * ITLE (Italic End)
	 * IMCB (Image Center Begin)
	 * IMRB (Image Right Begin)
	 * IMLB (Image Left Begin)
	 * DT (Draw Title) rectangular path (from DR), x, y, title, font size, font
	 * DHB (Draw Header Box) rectangular path (from DR), greyscale value (for filling), width, height, LLeft x, LLeft y (for clipping) 
	 * DUID (Draw User ID) x, y, title, font size, font (for title), width, height, LLeft x, LLeft y (for clipping) 
	 * DUP2 - Duplicate the top 2 items : x y -> x y x y
	 * EXCH2 - Exchange the next-to-top 2 items : a b c d -> b a c d
	 */
	fprintf( destFile, "/DUP2  { dup 3 2 roll dup 4 1 roll exch } bind def \n" );
	fprintf( destFile, "/EXCH2 { 4 2 roll exch 4 2 roll } bind def \n" );
	fprintf( destFile, "/RL    { 3 1 roll } bind def \n");
	fprintf( destFile, "/CP    { currentpoint } bind def \n");
	fprintf( destFile, "/DR    { moveto 1 index 0 rlineto 0 exch rlineto neg 0 rlineto closepath } bind def \n");
	fprintf( destFile, "/DL    { moveto lineto stroke } bind def \n");
	fprintf( destFile, "/SF    { findfont exch dup /CFS exch def scalefont setfont } bind def \n");
	fprintf( destFile, "/DT    { gsave DR clip newpath SF CS grestore } bind def \n");
	fprintf( destFile, "/ULB   { CP } bind def \n");
	fprintf( destFile, "/ULE   { gsave 1 sub moveto 1 sub lineto .25 setlinewidth stroke grestore } bind def \n");
	fprintf( destFile, "/DHB   { gsave setgray fill grestore clip stroke } bind def \n");
	fprintf( destFile, "/DTP   { gsave DR clip newpath SF RJ grestore } bind def \n");
	fprintf( destFile, "/SLN   { gsave DR initclip clip newpath LJ grestore } bind def \n");
	fprintf( destFile, "/BLDB  { CP gsave %d /%s SF } bind def \n", d_psProps->bodyFontSizePts, d_psProps->boldFont);
	fprintf( destFile, "/BLDE  { CP grestore moveto } bind def \n");
	fprintf( destFile, "/ITLB  { CP gsave %d /%s SF } bind def \n", d_psProps->bodyFontSizePts, d_psProps->italicFont);
	fprintf( destFile, "/ITLE  { CP grestore moveto } bind def \n");
	fprintf( destFile, "/DUID  { gsave DR clip newpath SF LJ grestore } bind def \n");
	fprintf( destFile, "/GRID  { gsave 1 setlinewidth 0 setgray [3] 0 setdash 36 36 612 { 0 moveto 0 792 rlineto } for \n");
	fprintf( destFile, "         36 36 792 { 0 exch moveto 612 0 rlineto } for stroke grestore } bind def \n");
	fprintf( destFile, "/CROSS { gsave 1 setlinewidth 306 0 moveto 0 792 rlineto stroke 0 396 moveto 612 0 rlineto stroke grestore } bind def \n");

	fprintf( destFile, "/IMCB  { CP 8 2 roll gsave exch RL exch dup 5 1 roll sub EXCH2 translate DUP2 0 0 DR clip scale mark } bind def\n" );
	fprintf( destFile, "/IMLB  { CP 8 2 roll gsave exch RL exch dup 5 1 roll sub EXCH2 translate DUP2 0 0 DR clip scale mark } bind def\n" );
	fprintf( destFile, "/IMRB  { CP 8 2 roll gsave exch RL exch dup 5 1 roll sub EXCH2 translate DUP2 0 0 DR clip scale mark } bind def\n" );

	fprintf( destFile, "/IMCE  { grestore cleartomark exch pop sub 2 sub moveto } bind def\n" );
	fprintf( destFile, "/IMRE  { grestore cleartomark exch pop sub 2 sub moveto } bind def\n" );
	fprintf( destFile, "/IMLE  { grestore cleartomark exch pop sub 2 sub moveto } bind def\n" );

	fprintf( destFile, "\n");
	fprintf( destFile, "%%%%EndProlog\n");
	fprintf( destFile, "\n");

	if( ferror(destFile) ) g_errorType = XnPS_WRITE_ERROR;
}

/*******************************************************/

/*
 *  Proc:	PrintCredits
 *  Callers:	PrintTrailer, EndPage
 *  Purpose:	Print some credits at the bottom left of the page.
 *  Params:	destFile - The PostScript output file.
 *		d_physicalPageNum - The physical page number this is.
 *		d_psProps - The PostScript properties.
 *  Globals:	g_errorType
 *  Returns:	None
 */
void PrintCredits( FILE *destFile, int d_physicalPageNum, XnPSPropsStruct *d_psProps )
{
	if( !d_psProps->printCredits ) return;

	switch( d_psProps->sheetFormat )
	{
	case XnPS_PORTRAIT:
		if_fprintf( destFile, "    gsave 8 /%s SF (NXpr v%s) %d %d RJ grestore\n",
				d_psProps->bodyFont, VERSION_STR, d_psProps->physicalPageWidthPts - d_psProps->rightMarginPts, BOTTOM_THRESHOLD - 10 );
		break;

	case XnPS_FOUR_ONE:
		if_fprintf( destFile, "    gsave 8 /%s SF (NXpr v%s) %d %d RJ grestore\n",
				d_psProps->bodyFont, VERSION_STR, d_psProps->physicalPageWidthPts - d_psProps->rightMarginPts, BOTTOM_THRESHOLD - 10 );
		break;

	case XnPS_LANDSCAPE:
		if_fprintf( destFile, "    gsave %d 0 translate 90 rotate 8 /%s SF (NXpr v%s) %d %d RJ grestore\n",
				d_psProps->physicalPageWidthPts,
				d_psProps->bodyFont, VERSION_STR, d_psProps->physicalPageHeightPts - d_psProps->rightMarginPts, BOTTOM_THRESHOLD - 10 );
		break;

	case XnPS_TWO_ONE:
		if_fprintf( destFile, "    gsave %d 0 translate 90 rotate 8 /%s SF (NXpr v%s) %d %d RJ grestore\n",
				d_psProps->physicalPageWidthPts,
				d_psProps->bodyFont, VERSION_STR, d_psProps->physicalPageHeightPts - d_psProps->rightMarginPts, BOTTOM_THRESHOLD - 10 );
		break;

	}

	if( ferror(destFile) ) g_errorType = XnPS_WRITE_ERROR;
}

/*******************************************************/

/*
 *  Proc:	PrintTrailer
 *  Callers:	XnGeneratePostScript
 *  Purpose:	Output the PostScript file trailer.
 *  Params:	destFile - The PostScript output file.
 *		d_physicalPageNum - The physical page number this is.
 *		virtualPageNum - The virtual page number this is.
 *		pPageNum - ?
 *		d_psProps - The PostScript properties.
 *		newPhysicalPageNum - <Return> Changes if we need to start another physical page.
 *  Globals:	g_errorType, g_showPageCnt
 *  Returns:	None
 */
void PrintTrailer( FILE *destFile, int d_physicalPageNum, int virtualPageNum, int pPageNum,
			XnPSPropsStruct *d_psProps, int *newPhysicalPageNum )
{
	switch( d_psProps->sheetFormat )
	{
	case XnPS_TWO_ONE:
		if( (virtualPageNum % 2) != 0 )
		{
			if( PAGE_RANGE ) g_showPageCnt++;
			PrintCredits( destFile, d_physicalPageNum, d_psProps );
			if_fprintf( destFile, "showpage\n" );
			pPageNum++;
		}
		break;

	case XnPS_FOUR_ONE:
		if( (virtualPageNum % 4) != 0 )
		{
			if( PAGE_RANGE ) g_showPageCnt++;
			PrintCredits( destFile, d_physicalPageNum, d_psProps );
			if_fprintf( destFile, "showpage\n" );
			pPageNum++;
		}
		break;
	}

	fprintf( destFile, "%%%%Trailer\n");
	fprintf( destFile, "%%%%Pages: %d\n", g_showPageCnt );
	fprintf( destFile, "%%%%EOF\n");

	(*newPhysicalPageNum) = pPageNum;

	if( ferror(destFile) ) g_errorType = XnPS_WRITE_ERROR;
}

/***********************************************************/

/*
 *  Proc:	DrawHeader
 *  Callers:	BeginNewPagePS
 *  Purpose:	Draw the sections that should be included in the header.
 *  Params:	destFile - The destination output file.
 *		virtualPageNum - The virtual page number.
 *		d_physicalPageNum - The physical page number (for macros).
 *		d_psProps - The PS output properties.
 *  Globals:	g_startingPageNum
 *  Returns:	None
 */
void DrawHeader( FILE *destFile, int virtualPageNum, int d_physicalPageNum, XnPSPropsStruct *d_psProps )
{
	int	lineY		= 0;
	int	titleY		= 0;
	char	s[ MAX_LINE_LEN ];
	char	pageString[ MAX_LINE_LEN ];
	Boolean	output		= False;

	if( d_psProps->allPages ) output = True;
	else
	{
		switch( d_psProps->sheetFormat )
		{
		case XnPS_PORTRAIT:
		case XnPS_LANDSCAPE:
			if( virtualPageNum == g_startingPageNum ) output = True;
			else output = False;
			break;

		case XnPS_TWO_ONE:
			if( virtualPageNum == ((g_startingPageNum * 2) - 1) ) output = True;
			else output = False;
			break;

		case XnPS_FOUR_ONE:
			if( virtualPageNum == ((g_startingPageNum * 4) - 3) ) output = True;
			else output = False;
			break;
		}
	}

	if( (output == False) && (d_psProps->printPageNumbers == False) ) return;

	if_fprintf( destFile, "    gsave    %% Begin Header\n" );

	/*
	 * Draw the shaded box at the top of the page for the header information.
	 * Clip the header info into this box so that it dosen't look sloppy.
	 */
	lineY = TOP_THRESHOLD - HEADER_THICKNESS;
	titleY = lineY + d_psProps->headerLineTopOffsetPts + I_HALF(d_psProps->titleFontSize);

	if_fprintf( destFile, "        %.2f setlinewidth %d %d %d %d %s %.2f %s  %% HB\n",
			d_psProps->borderLineWidth,

			/* width, height, llx, lly path for clipping. */
			d_psProps->virtualPageWidthPts,
			HEADER_THICKNESS + d_psProps->topOffsetPts,
			d_psProps->leftMarginPts,
			lineY,
			DRAW_RECTANGLE_PROC,

			/* grey value for filling. */
			d_psProps->headerBackground,
			DRAW_HEADER_BOX_PROC );

	/*
	 * Take a look at the things that are printed on the right side
 	 * of the header - the page numbers and time.
	 */
	if( (d_psProps->printPageNumbers) || (d_psProps->printTime) )
	{
		if( d_psProps->printTotalPages )
			sprintf( pageString, "%d/%d", virtualPageNum, d_psProps->maxPageNumber );
		else
			sprintf( pageString, "Page %2d", virtualPageNum );

		if( output && (d_psProps->printTime) && (!d_psProps->printPageNumbers) )
			sprintf( s, "%s", d_psProps->time );
		else
		{
			if( output && (!d_psProps->printTime) && (d_psProps->printPageNumbers) )
				sprintf( s, "%s", pageString );
			else
			{
				if( output && (d_psProps->printTime) && (d_psProps->printPageNumbers) )
					sprintf( s, "%s   %s", d_psProps->time, pageString );
				else
				{
					if( (!output) && (d_psProps->printPageNumbers) )
						sprintf( s, "%s", pageString );
				}
			}
		}

		if_fprintf( destFile, "        (%s) %d %d %d /%s %d %d %d %d %s\n",

				/* page string, x, y, font size, font name. */
				s,
				RIGHT_THRESHOLD - 1,
				titleY - I_HALF(d_psProps->headerFontSize) + I_QUART(d_psProps->headerFontSize),
				d_psProps->headerFontSize,
				d_psProps->headerFont,

				/* width, height, llx, lly for clipping page area. */
				TIME_PAGE_WIDTH_PTS,
				HEADER_THICKNESS,
				RIGHT_THRESHOLD - TIME_PAGE_WIDTH_PTS,
				lineY,
				DRAW_TIME_PAGE_PROC );
	}

	/*
	 * Take a look at the info that is put in the center of the header.
	 * The title...
	 */
	if( d_psProps->printTitle && output )
	{
		if_fprintf(destFile, "        (%s) %d %d %d /%s %d %d %d %d %s\n",

				/* title, x, y, font size, font name. */
				d_psProps->title,
				d_psProps->leftMarginPts + I_HALF(d_psProps->virtualPageWidthPts),
				titleY - I_HALF(d_psProps->titleFontSize) + I_QUART(d_psProps->titleFontSize),
				d_psProps->titleFontSize,
				d_psProps->titleFont,

				/* width, height, llx, lly for clipping title area. */
				TITLE_WIDTH_PTS,
				HEADER_THICKNESS,
				d_psProps->leftMarginPts + I_HALF(d_psProps->virtualPageWidthPts) - I_HALF(TITLE_WIDTH_PTS),
				lineY,
				DRAW_TITLE_PROC);
	}

	/*
	 * Take a look at the info that is put of the left side of the header.
	 * The user's ID...
	 */
	if( d_psProps->printUserID && output )
	{
		if_fprintf( destFile, "        (%s) %d %d %d /%s %d %d %d %d %s\n",

				/* user Id, x, y, font size, font name. */
				d_psProps->userID,
				HEADER_LEFT_THRESHOLD,
				titleY - I_HALF(d_psProps->headerFontSize) + I_QUART(d_psProps->headerFontSize),
				d_psProps->headerFontSize,
				d_psProps->headerFont,

				/* width, height, llx, lly for clipping id area. */
				USER_ID_WIDTH_PTS,
				HEADER_THICKNESS,
				HEADER_LEFT_THRESHOLD,
				lineY,
				DRAW_USERID_PROC );
	}

	if_fprintf( destFile, "    grestore    %% End Header\n" );


	/*
	 * Print a border around the header area.  Didn't do this earlier because we
	 * want to be outside the clip area.
	 */
	if( d_psProps->printBorder )
	{
		if_fprintf( destFile, "    %d %d %d %d %.2f setlinewidth %s  %% HL\n", 
				d_psProps->leftMarginPts,
				lineY,
				d_psProps->leftMarginPts + d_psProps->virtualPageWidthPts,
				lineY,
				d_psProps->headerLineWidth,
				DRAW_LINE_PROC );
	}
}

/***********************************************************/

/*
 *  Proc:	BeginNewPagePS
 *  Callers:	XnGeneratePostScript
 *  Purpose:	Output the PostScript that is required to start a new page.
 *  Params:	destFile - The destination output file.
 *		virtualPageNum - The virtual page number.
 *		d_physicalPageNum - The physical page number (for macros).
 *		d_psProps - The PS output properties.
 *  Globals:	g_errorType, g_startingPageNum
 *  Returns:	None
 */
void BeginNewPagePS( FILE *destFile, int virtualPageNum, int d_physicalPageNum, XnPSPropsStruct *d_psProps )
{
	/*
	 * Depending on what sheet format we are using set up the
	 * page orientation, scale, translations and any other initializing
	 * that gets done on a per sheet basis.
	 */
	switch( d_psProps->sheetFormat )
	{
	case XnPS_PORTRAIT:
		if_fprintf( destFile, "\n%%%%Page: %d %d\n", d_physicalPageNum - g_startingPageNum + 1, d_physicalPageNum - g_startingPageNum + 1 );
		if_fprintf( destFile, "gsave\n" );
		if_fprintf( destFile, "    %f %f scale\n", d_psProps->xScale, d_psProps->yScale );
		break;

	case XnPS_LANDSCAPE:
		if_fprintf( destFile, "\n%%%%Page: %d %d\n", d_physicalPageNum - g_startingPageNum + 1, d_physicalPageNum - g_startingPageNum + 1 );
		if_fprintf( destFile, "gsave\n" );
		if_fprintf( destFile, "    %d 0 translate 90 rotate\n", d_psProps->physicalPageWidthPts );
		if_fprintf( destFile, "    %f %f scale\n", d_psProps->xScale, d_psProps->yScale );
		break;


	/*
	 *  Where the virtual sheet is translated to...
	 *  The sheet on the left is translated x to the left margin and y to the bottom margin.
	 *  The subtraction of the translate is because the final sheet is printed with a left
	 *  and bottom margin both of which are scaled.  We must offset this or the sheets
	 *  will have a left and bottom margin that are twice what they should be.
	 *  The sheet on the right has the same translation except that it's left side will
	 *  be offset half the PHYSICAL page HEIGHT plus half the inner margin.
	 */
	case XnPS_TWO_ONE:
		if( (virtualPageNum % 2) == 1 )
			if_fprintf( destFile, "\n%%%%Page: %d %d\n",
				d_physicalPageNum - g_startingPageNum + 1, d_physicalPageNum - g_startingPageNum + 1 );

		if_fprintf( destFile, "gsave    %% Start Sheet %d\n", virtualPageNum );
		if_fprintf( destFile, "    %d 0 translate 90 rotate\n", d_psProps->physicalPageWidthPts );

		/* 1 = Left side ;  2 = Right side */
		switch( virtualPageNum % 2 )
		{
		case 1:
			if_fprintf( destFile, "    %d %d translate\n",
					d_psProps->leftMarginPts - (int)(d_psProps->leftMarginPts * d_psProps->xScale),
					d_psProps->bottomMarginPts - (int)(d_psProps->bottomMarginPts * d_psProps->yScale) );
			break;

		case 0:
			if_fprintf( destFile, "    %d %d translate\n",
					I_HALF(d_psProps->physicalPageHeightPts) + I_HALF(d_psProps->innerMarginLRPts) -
					    (int)(d_psProps->leftMarginPts * d_psProps->xScale),
					d_psProps->bottomMarginPts - (int)(d_psProps->bottomMarginPts * d_psProps->yScale) );
			break;
		}

		if_fprintf( destFile, "    %f %f scale\n", d_psProps->xScale, d_psProps->yScale );
		break;

	case XnPS_FOUR_ONE:
		if( (virtualPageNum % 4) == 1 )
			if_fprintf( destFile, "\n%%%%Page: %d %d\n",
				d_physicalPageNum - g_startingPageNum + 1, d_physicalPageNum - g_startingPageNum + 1 );

		if_fprintf( destFile, "gsave    %% Start Sheet %d\n", virtualPageNum );

		/* 1 = UL ; 2 = UR ; 3 = LL ; 0 = LR  (U=Upper, L=Left / Lower, R=Right). */
		switch( virtualPageNum % 4 )
		{
		case 1:
			if_fprintf( destFile, "    %d %d translate\n",
				d_psProps->leftMarginPts - (int)(d_psProps->leftMarginPts * d_psProps->xScale),
				I_HALF(d_psProps->physicalPageHeightPts) + I_HALF(d_psProps->innerMarginTBPts) -
				    (int)(d_psProps->bottomMarginPts * d_psProps->yScale) );
			break;

		case 2:
			if_fprintf( destFile, "    %d %d translate\n",
				I_HALF(d_psProps->physicalPageWidthPts) + I_HALF(d_psProps->innerMarginLRPts) -
				    (int)(d_psProps->leftMarginPts * d_psProps->xScale),
				I_HALF(d_psProps->physicalPageHeightPts) + I_HALF(d_psProps->innerMarginTBPts) -
				    (int)(d_psProps->bottomMarginPts * d_psProps->yScale) );
			break;

		case 3:
			if_fprintf( destFile, "    %d %d translate\n",
				d_psProps->leftMarginPts - (int)(d_psProps->leftMarginPts * d_psProps->xScale),
				d_psProps->bottomMarginPts - (int)(d_psProps->bottomMarginPts * d_psProps->yScale) );
			break;

		case 0:
			if_fprintf( destFile, "    %d %d translate\n",
				I_HALF(d_psProps->physicalPageWidthPts) + I_HALF(d_psProps->innerMarginLRPts) -
				    (int)(d_psProps->leftMarginPts * d_psProps->xScale),
				d_psProps->bottomMarginPts - (int)(d_psProps->bottomMarginPts * d_psProps->yScale) );
			break;
		}

		if_fprintf( destFile, "    %f %f scale\n", d_psProps->xScale, d_psProps->yScale );
		break;
	}

	/*
	 * Now that we have the page orientation set up, we need to add on
	 * other things that get done for each page such as headers, borders, titles, etc...
	 */

	/* Draw a border around the virtual page. */
	if( d_psProps->printBorder )
		if_fprintf( destFile, "    %.2f setlinewidth %d %d %d %d %s stroke\n",
				d_psProps->borderLineWidth,

				/* width, height, llx, lly for drawing border. */
				d_psProps->virtualPageWidthPts,
				d_psProps->virtualPageHeightPts,
				d_psProps->leftMarginPts,
				d_psProps->bottomMarginPts,
				DRAW_RECTANGLE_PROC );

	/* Print a line between the line numbers and text. */
	if( d_psProps->printLineNumbers )
	{
		if_fprintf( destFile, "    %.2f setlinewidth %d %d %d %d %s stroke  %% LN\n",
				d_psProps->borderLineWidth,

				/* x1, y1, x2, y2 of line. */
				BODY_LEFT_THRESHOLD - NUMBERING_LINE_OFFSET, 
				d_psProps->bottomMarginPts,
				BODY_LEFT_THRESHOLD - NUMBERING_LINE_OFFSET, 
				TOP_THRESHOLD,
				DRAW_LINE_PROC );
	}

	/* If we need the header area, draw a dividing line between the header and text. */
	if( d_psProps->needHeaderArea ) DrawHeader(destFile, virtualPageNum, d_physicalPageNum, d_psProps);

	/* Set the clipping rectangle for the body to be a little smaller than the virtual page size. */
	if_fprintf( destFile, "    %.2f setlinewidth %d %d %d %d %s clip newpath\n",
			d_psProps->borderLineWidth,

			/* width, height, llx, lly for clipping text. */
			d_psProps->virtualPageWidthPts - (d_psProps->bodyLeftOffsetPts + d_psProps->bodyRightOffsetPts),
			d_psProps->virtualPageHeightPts - (d_psProps->topOffsetPts + d_psProps->bottomOffsetPts),
			BODY_LEFT_THRESHOLD,
			BOTTOM_THRESHOLD,
			DRAW_RECTANGLE_PROC );

	/* Set the body font size. */
	if_fprintf( destFile, "    %d /%s %s\n", d_psProps->bodyFontSizePts, d_psProps->bodyFont, SET_FONT_PROC );

	if( ferror(destFile) ) g_errorType = XnPS_WRITE_ERROR;
}

/***********************************************************/

/*
 *  Proc:	EndPage
 *  Callers:	XnGeneratePostScript
 *  Purpose:	End the current page with a 'showpage'.
 *  Params:	destFile - The destination output file.
 *		virtualPageNum - The current virtual page number.
 *		d_physicalPageNum - The physical page number (for macros).
 *		d_psProps - The PS output properties.
 *		newVirtualPageNum - The new virtual page number.
 *		newPhysicalPageNum - The new physical page number.
 *  Globals:	g_errorType, g_showPageCnt
 *  Returns:	None
 */
void EndPage( FILE *destFile, int virtualPageNum, int d_physicalPageNum,
		XnPSPropsStruct *d_psProps, int *newVirtualPageNum, int *newPhysicalPageNum )
{
	(*newVirtualPageNum) = virtualPageNum + 1;

	/* 
	 * If the next sheet to be printed is going to cause a new page to be used,
	 * increment the newPhysicalPageNum, otherwise the next sheet will be on
	 * the current page.
	 */
	switch( d_psProps->sheetFormat )
	{
	case XnPS_TWO_ONE:
		if( (virtualPageNum % 2) == 0 )
		{
			(*newPhysicalPageNum) = d_physicalPageNum + 1;

			if( PAGE_RANGE ) g_showPageCnt++;
			if_fprintf(destFile, "    grestore\n");
			PrintCredits( destFile, d_physicalPageNum, d_psProps );
			if_fprintf(destFile, "showpage\n");
		}
		else if_fprintf(destFile, "    grestore\n");

		break;

	case XnPS_FOUR_ONE:
		if( (virtualPageNum % 4) == 0 )
		{
			(*newPhysicalPageNum) = d_physicalPageNum + 1;

			if( PAGE_RANGE ) g_showPageCnt++;
			if_fprintf(destFile, "    grestore\n");
			PrintCredits( destFile, d_physicalPageNum, d_psProps );
			if_fprintf(destFile, "showpage\n");
		}
		else if_fprintf(destFile, "    grestore\n");

		break;

	default:
		(*newPhysicalPageNum) = d_physicalPageNum + 1;

		if( PAGE_RANGE ) g_showPageCnt++;
		if_fprintf(destFile, "    grestore\n");
		PrintCredits( destFile, d_physicalPageNum, d_psProps );
		if_fprintf(destFile, "showpage\n");

		break;
	}

	if( ferror(destFile) ) g_errorType = XnPS_WRITE_ERROR;
}

/***********************************************************/

/*
 *  Proc:	GetInputLine
 *  Callers:	XnGeneratePostScript
 *  Purpose:	Read a line from the input file are return it.
 *  Params:	srcFile - The input file.
 *		srcLineChars - <Return> The number of characters in the line.
 *  Globals:	g_errorType
 *  Returns:	The line read.
 */
char *GetInputLine( FILE *srcFile, int *srcLineChars, int *linesProcessed )
{
	int	len;
	char	str[ MAX_LINE_LEN ];

	(*srcLineChars) = 0;

	if( fgets( str, MAX_LINE_LEN, srcFile ) != NULL )
	{
		(*linesProcessed)++;

		str[ MAX_LINE_LEN  - 1 ] = '\0';

		len = strlen( str );
		if( str[len - 1] == '\n' ) str[ len - 1 ] = '\0';

		(*srcLineChars) = strlen( str );

		return( XtNewString( str ) );
	}

	if( ferror(srcFile) ) g_errorType = XnPS_READ_ERROR;

	return( NULL );
}

/***********************************************************/

/*
 *  Proc:	EndEscapedStartNormalText
 *  Callers:	ProcessEscapeCharacter, ProcessInputLine
 *  Purpose:	End the currently escaped text style and start a normal text style.
 *  Params:	currentEscapeType - The current escape type to end.
 *		destFile - The destination output file.
 *		d_psProps - The PS output properties.
 *		d_physicalPageNum - The physical page number (for macros).
 *  Globals:	None
 *  Returns:	None
 */
void EndEscapedStartNormalText( int currentEscapeType, FILE *destFile, XnPSPropsStruct *d_psProps, int d_physicalPageNum )
{
	switch( currentEscapeType )
	{
	case TEXT_UNDERLINE:
		if_fprintf( destFile, ") CP %s CP ULE\n", SHOW_LINE_PROC );
		break;

	case TEXT_BOLD:
		if_fprintf( destFile, ") RL %s BLDE\n", SHOW_LINE_PROC );
		break;

	case TEXT_ITALIC:
		if_fprintf( destFile, ") RL %s ITLE\n", SHOW_LINE_PROC );
		break;

	case IMAGE_CENTER:
		if_fprintf( destFile, "    IMCE\n" );
		break;

	case IMAGE_LEFT:
		if_fprintf( destFile, "    IMLE\n" );
		break;

	case IMAGE_RIGHT:
		if_fprintf( destFile, "    IMRE\n" ); 
		break;

	}

	if_fprintf(destFile, "    (");
}

/***********************************************************/

/*
 *  Proc:	EndLine
 *  Callers:	EndNormalStartEscapedLine, ProcessInputLine
 *  Purpose:	End the current escaped line style.
 *  Params:	destFile - The destination output file.
 *		notCurrentlyEscaped - True if we are not already in escaped mode.
 *		lineX - The x-coordinate for the PostScript line.
 *		lineY - The y-coordinate for the PostScript line.
 *		d_physicalPageNum - The physical page number (for macros).
 *		d_psProps - The PS output properties.
 *  Globals:	None
 *  Returns:	None
 */
void EndLine( FILE *destFile, Boolean notCurrentlyEscaped, int lineX, int lineY, int d_physicalPageNum, XnPSPropsStruct *d_psProps )
{
	if( notCurrentlyEscaped )
	{
		if_fprintf( destFile, ") %d %d %s\n", lineX, lineY, SHOW_LINE_PROC );
	}
	else
	{
		if_fprintf( destFile, ") CP %s\n", SHOW_LINE_PROC );
	}
}

/***********************************************************/

/*
 *  Proc:	EndNormalStartEscapedLine
 *  Callers:	ProcessEscapeCharacter, ProcessInputLine
 *  Purpose:	End the current processed text style and start a new style.
 *  Params:	currentEscapeType - What type of processing to start.
 *		notCurrentlyEscaped - True if we are not currently escaped.
 *		destFile - The destination output file.
 *		d_physicalPageNum - The physical page number (for macros).
 *		lineX - The x-coordinate for the PostScript line.
 *		lineY - The y-coordinate for the PostScript line.
 *		d_psProps - The PS output properties.
 *  Globals:	None
 *  Returns:	None
 */
void EndNormalStartEscapedLine( int currentEscapeType, Boolean notCurrentlyEscaped, FILE *destFile,
			int d_physicalPageNum, int lineX, int lineY,
			XnPSPropsStruct *d_psProps )
{
	switch( currentEscapeType )
	{
	case TEXT_UNDERLINE:

		EndLine( destFile, notCurrentlyEscaped, lineX, lineY, d_physicalPageNum, d_psProps );
		if_fprintf( destFile, "    ULB (" );

		break;

	case TEXT_BOLD:

		EndLine( destFile, notCurrentlyEscaped, lineX, lineY, d_physicalPageNum, d_psProps );
		if_fprintf( destFile, "    BLDB (" );

		break;

	case TEXT_ITALIC:

		EndLine( destFile, notCurrentlyEscaped, lineX, lineY, d_physicalPageNum, d_psProps );
		if_fprintf( destFile, "    ITLB (" );

		break;

	case IMAGE_CENTER:
	case IMAGE_LEFT:
	case IMAGE_RIGHT:

		EndLine( destFile, notCurrentlyEscaped, lineX, lineY, d_physicalPageNum, d_psProps );

		break;

	default:
		break;
	}
}

/***********************************************************/

/*
 *  Proc:	ProcessEscapeCharacter
 *  Callers:	ProcessInputLine
 *  Purpose:	Process an escape character.  If the escape is not NORMAL_TEXT then
 *		end a line of "normal" text and start a line of escaped text.  If
 *		the escape is a NORMAL_TEXT escape, end the escaped text and start
 *		a line of normal text.
 *  Params:	notCurrentlyEscaped - True if the currently processed line is NORMAL.
 *		destFile - The destination PostScript file.
 *		srcLine - The currently processed line.
 *		d_physicalPageNum - The 
 *		lineX, lineY - The coordinates for the currently processed line.
 *		d_psProps - The PostScript file properties.
 *		srcLineIndex - <Return> The index of the currently processed character.
 *		currentEscapeType - <Return> The type of escape processed.
 *  Globals:	None
 *  Returns:	None
 */
void ProcessEscapeCharacter( Boolean notCurrentlyEscaped, FILE *destFile, char *srcLine,
			int d_physicalPageNum, int lineX, int lineY,
			XnPSPropsStruct *d_psProps, int *srcLineIndex, int *currentEscapeType )
{
	int	newIndex;
	Boolean	duplicate;

	/*
	 * If we hit a control character and it's not a VT100 code
	 * (i.e. dosen't have a [ following the control) then skip
	 * the control character and move on.
	 */
	if( srcLine[ (*srcLineIndex) + 1 ] != '[' )
	{
		(*srcLineIndex)++;
		return;
	}

	duplicate = False;
	newIndex = (*srcLineIndex) + 2;

	/*
	 * Find out what type of escape character is being processed.
	 * If the escape character is already being processed, ignore the
	 * second occurence of the control.
	 */
	switch( srcLine[newIndex] )
	{
	case IMAGE_CENTER_ESCAPE:
		if( srcLine[ newIndex + 1 ] != 'm' ) break;
		newIndex += 2;

		if( (*currentEscapeType) == IMAGE_CENTER ) duplicate = True;
		else
		{
			(*currentEscapeType) = IMAGE_CENTER;
		}

		break;

	case IMAGE_LEFT_ESCAPE:
		if( srcLine[ newIndex + 1 ] != 'm' ) break;
		newIndex += 2;

		if( (*currentEscapeType) == IMAGE_LEFT ) duplicate = True;
		else
		{
			(*currentEscapeType) = IMAGE_LEFT;
		}

		break;

	case IMAGE_RIGHT_ESCAPE:
		if( srcLine[ newIndex + 1 ] != 'm' ) break;
		newIndex += 2;

		if( (*currentEscapeType) == IMAGE_RIGHT ) duplicate = True;
		else
		{
			(*currentEscapeType) = IMAGE_RIGHT;
		}

		break;

	case TEXT_UNDERLINE_ESCAPE:
		if( srcLine[ newIndex + 1 ] != 'm' ) break;
		newIndex += 2;

		if( (*currentEscapeType) == TEXT_UNDERLINE ) duplicate = True;
		else
		{
			(*currentEscapeType) = TEXT_UNDERLINE;
		}

		break;

	case TEXT_BOLD_ESCAPE:
		if( srcLine[ newIndex + 1 ] != 'm' ) break;
		newIndex += 2;

		if( (*currentEscapeType) == TEXT_BOLD ) duplicate = True;
		else
		{
			(*currentEscapeType) = TEXT_BOLD;
		}

		break;

	case TEXT_ITALIC_ESCAPE:
		if( srcLine[ newIndex + 1 ] != 'm' ) break;
		newIndex += 2;

		if( (*currentEscapeType) == TEXT_ITALIC ) duplicate = True;
		else
		{
			(*currentEscapeType) = TEXT_ITALIC;
		}

		break;

	case TEXT_NORMAL_ESCAPE:
		if( (*currentEscapeType) != TEXT_NORMAL )
		{
			EndEscapedStartNormalText( (*currentEscapeType), destFile, d_psProps, d_physicalPageNum );
			(*currentEscapeType) = TEXT_NORMAL;
			newIndex += 1;
		}
		break;

	default:
		break;

	}

	/*
 	 * If we started a new escape type we need to finish up the current
	 * text and set things up for the new text style.
	 */
	if( ((*currentEscapeType) != TEXT_NORMAL) && !duplicate )
	{
		EndNormalStartEscapedLine( (*currentEscapeType), notCurrentlyEscaped, destFile, d_physicalPageNum, lineX, lineY, d_psProps );
	}

	(*srcLineIndex) = newIndex;
}

/***********************************************************/

/*
 *  Proc:	StartProcessingImage
 *  Callers:	ProcessInputLine
 *  Purpose:	Starts the processing of an inline PostScript image that is embedded between
 *		some control characters.
 *  Params:	destFile - The destination output file.
 *		srcLine - The line of text that starts the image.
 *		srcLineIndex - The index into the srcLine that starts the image.
 *		d_physicalPageNum - The physical page number (for macros).
 *		d_psProps - The PS output properties.
 *		currentEscapeType - The image escape type.
 *		lineY - The y-coordinate of the current line in PostScript we are on.
 *		newLineY - <Return> What the new lineY should be after processing the image.
 *  Globals:	g_imageCnt
 *  Returns:	Any text that needs to be wrapped to the next line.  NULL if the image will
 *		fit on the current page.
 */
char *StartProcessingImage( FILE *destFile, char *srcLine, int srcLineIndex,
				int d_physicalPageNum, XnPSPropsStruct *d_psProps, int currentEscapeType,
				int lineY, int *newLineY )
{
	int		imageY;
	int		bboxWidth;
	int		bboxHeight;
	int		fontOffset;
	char		*s			=NULL;
	char		*t			=NULL;
	char		tmp[ MAX_LINE_LEN ];
	char		*excessSrcLine		=NULL;

	/*
	 * Try to read the first input line in processing the image.
	 * This line SHOULD contain a bounding box width/height for the image.
	 * This line format was output with the XnPixmapToPostScript library call.
	 */
	strcpy( tmp, &srcLine[srcLineIndex] );
	s = strtok( tmp, " " );
	t = strtok( NULL, " " );

	if( s && t )
	{
		errno = 0;
		bboxWidth = atoi( s );
		bboxHeight = atoi( t );

		/*
		 * Convert the text into an integer width/height and make sure that
		 * the image will fit on the current page.  If it won't fit on the page
		 * and must be wrapped, return the text to the caller and process it later.
		 * If it will fit on the current page, move the lineY down - this is where
		 * the next line of text will appear after the image.
		 */
		if( errno != ERANGE )
		{
			fontOffset = BODY_LINE_HEIGHT - 2;

			imageY = lineY + fontOffset;
			(*newLineY) = (lineY - bboxHeight - BODY_LINE_HEIGHT) + fontOffset;

			/* It won't fit. */
			if( ((*newLineY) + BODY_LINE_HEIGHT) < BOTTOM_THRESHOLD )
			{
				excessSrcLine = XnMallocString( strlen( &(srcLine[srcLineIndex - 4]) ) + 1 );
				strcpy( excessSrcLine, &(srcLine[srcLineIndex - 4]) );
				return( excessSrcLine );
			}
			else
			{
				/* What type of image processing was requested? */
				switch( currentEscapeType )
				{
					case IMAGE_CENTER:
						g_imageCnt++;
						if_fprintf( destFile, "    /IMCB %d %d %d %d %d %d 7 6 roll cvx exec\n",
								bboxWidth, bboxHeight, bboxWidth, bboxHeight,
								d_psProps->leftMarginPts + (I_HALF(d_psProps->virtualPageWidthPts) - ( bboxWidth / 2)),
								imageY );

					break;

					case IMAGE_LEFT:
						g_imageCnt++;
						if_fprintf( destFile, "    /IMCL %d %d %d %d %d %d 7 6 roll cvx exec\n",
								bboxWidth, bboxHeight, bboxWidth, bboxHeight,
								BODY_LEFT_THRESHOLD, imageY );
					break;

					case IMAGE_RIGHT:
						g_imageCnt++;
						if_fprintf( destFile, "    /IMCR %d %d %d %d %d %d 7 6 roll cvx exec\n",
							bboxWidth, bboxHeight, bboxWidth, bboxHeight,
								( RIGHT_THRESHOLD -  bboxWidth ), imageY );
					break;
				}
			}
		}
	}

	return( NULL );
}

/***********************************************************/

/*
 *  Proc:	ProcessInputLine
 *  Callers:	XnGeneratePostScript
 *  Purpose:	Output the PostScript for a line of input.
 *  Params:	destFile - The PostScript output file.
 *		srcLine - The input line to process.
 *		currentLineNumber - The line number or -1 if this line has no number.
 *		currentEscapeType - <Return> The current escape type.
 *		d_physicalPageNum - The physical page number this line is on.
 *		lineX, lineY - The coordinates where this line should be placed.
 *		d_psProps - The PostScript file properties structure.
 *		newLineX, newLineY - The new coordinates for the next line of text.
 *		numLinesWrapped - <Return> The number of lines that have been wrapped (total).
 *		actualBodyTextWidth - <Return> The actual width of the page for body text.
 *		longestLineWidth - <Return> The width (points) of the longest line (so far).
 *  Globals:	g_errorType
 *  Returns:	Any extra text that was not processed and still needs to be.
 */
char *ProcessInputLine( FILE *destFile, char *srcLine, int *currentEscapeType, int currentLineNumber, int d_physicalPageNum,
			int lineX, int lineY, XnPSPropsStruct *d_psProps,
			int *newLineX, int *newLineY, int *numLinesWrapped, int *actualBodyTextWidth,
			float *longestLineWidth )
{
	int		srcLineIndex		=0;
	int		currentColumn		=0;
	char		*excessSrcLine		=NULL;
	float		currentLineWidth	=0.0;
	float		currentCharWidth	=0.0;
	Boolean		notCurrentlyEscaped	=True;

	currentColumn = 1;

	(*newLineX) = lineX;
	(*newLineY) = lineY;

	/* Determine how much width there is for the body text. */
	(*actualBodyTextWidth) = d_psProps->virtualPageWidthPts - (d_psProps->bodyLeftOffsetPts + d_psProps->bodyRightOffsetPts);

        /*********************************************************************************
         *
         * First print out the line number for the line we are processing.
	 * If the current line number is -1 then this line is a continuation
	 * of the previous line.
         *
         *********************************************************************************/

	if( ! PROCESSING_IMAGE(*currentEscapeType) )
	{
		if( d_psProps->printLineNumbers && (currentLineNumber >= 0) )
		{
			if_fprintf(destFile, "    (%d) %d %d %d %d %d %d %s",

				/* line number, x, y to show number. */
				currentLineNumber,
				d_psProps->leftMarginPts + NUMBERING_LINE_OFFSET,
				lineY,

				/* width, height, llx, lly for clipping line number text. */
				d_psProps->bodyLeftOffsetPts - NUMBERING_LINE_OFFSET,
				d_psProps->bodyFontSizePts,
				d_psProps->leftMarginPts,
				lineY,
				SHOW_LINE_NUM_PROC );
		}

		/*
		 * First start this line of text.  Then if the text being processed is not
		 * NORMAL then immediately close the NORMAL text (empty) and start an escaped
		 * line of text.
		 */
		if_fprintf(destFile, "    (");

		if( (*currentEscapeType) != TEXT_NORMAL )
		{
			EndNormalStartEscapedLine( (*currentEscapeType), notCurrentlyEscaped, destFile,
							d_physicalPageNum, lineX, lineY, d_psProps );
			notCurrentlyEscaped = False;
		}
	}

        /*********************************************************************************
         *
         * Now process the line... Run through the line looking at each individual
	 * character.  If we see a special character - do a special thing, otherwise
	 * just spit the character into the output file.  Make sure we don't run off the
	 * bottom or the right of the page.  Check for line-wrapping and page-wrapping.
         *
         *********************************************************************************/

	while( srcLine[srcLineIndex] != '\0' )
	{
		if( ferror(destFile) )
		{
			g_errorType = XnPS_WRITE_ERROR;
			return( NULL );
		}

		if( ! PROCESSING_IMAGE(*currentEscapeType) )
		{
			currentCharWidth =  PSCHARWIDTH( srcLine[srcLineIndex] );

			if( currentCharWidth < 0.0 ) currentCharWidth = 0.0;
			(*longestLineWidth) = ( currentLineWidth > (*longestLineWidth) ) ? currentLineWidth : (*longestLineWidth);

			/*
			 * Check to see if the character currently being processed would cause us to
			 * run off the right of the page.  If it would then we need to put the character on the
			 * next line.  Note: If the user has turned off character wrapping, this will never
			 * be true!  The text will run off the right of the page, but it will just get
			 * clipped by the clipping operator.
			 */
			if( ((currentLineWidth + currentCharWidth) > (float)(*actualBodyTextWidth)) )
			{
				if( (*currentEscapeType) != TEXT_NORMAL )
					EndEscapedStartNormalText( (*currentEscapeType), destFile, d_psProps, d_physicalPageNum );
				EndLine( destFile, notCurrentlyEscaped, lineX, lineY, d_physicalPageNum, d_psProps );

				if( PAGE_RANGE ) (*numLinesWrapped)++;

				(*newLineX) = BODY_LEFT_THRESHOLD;
				(*newLineY) = lineY - BODY_LINE_HEIGHT;

				/*
				 * It might be the case that the line is wrapped, AND the newly wrapped line will
				 * fall on the next page.  Return this to our caller and let them worry about
				 * creating a new page.
				 */
				if( (*newLineY) < BOTTOM_THRESHOLD )
				{
					excessSrcLine = XnMallocString( strlen( &(srcLine[srcLineIndex]) ) + 1 );
					strcpy( excessSrcLine, &(srcLine[srcLineIndex]) );
					return( excessSrcLine );
				}

				lineY = (*newLineY);
				lineX = (*newLineX);
		
				/* Print out the leading spaces before the wrapped line. */
				if_fprintf( destFile, "    (%s", LEADING_SPACES );

				currentLineWidth =  PSCHARWIDTH( d_psProps->leadingSpaces[0] ) * (float)strlen( LEADING_SPACES );
				currentColumn = 1 + strlen( LEADING_SPACES );

				notCurrentlyEscaped = True;

				if( (*currentEscapeType) != TEXT_NORMAL )
				{
					EndNormalStartEscapedLine( (*currentEscapeType), notCurrentlyEscaped,
									destFile, d_physicalPageNum, lineX, lineY, d_psProps);
					notCurrentlyEscaped = False;
				}
			}
		}

		/*
		 * Find out what character we are looking at.  If it is a special character
		 * (e.g. a control character) we may need to switch states.  If we are
		 * processing an image (e.g. inlining the current input text) we ignore
		 * most everything except escapes - everything else just gets output
		 * like normal.
		 */
		switch( srcLine[ srcLineIndex ] )
		{
		case NEWLINE:
			/* Should never happen - they are stripped off when read from the file. */
			break;

		case ESCAPE:
			if( d_psProps->processEscape == True )
			{
				ProcessEscapeCharacter( notCurrentlyEscaped, destFile, srcLine,
							d_physicalPageNum, lineX, lineY, d_psProps,
							&srcLineIndex, currentEscapeType );

				/*
				 * If the current escape just processed caused us to start
				 * inlining an image then start the processing of the image.
				 */
				if( PROCESSING_IMAGE(*currentEscapeType) )
				{
					excessSrcLine = StartProcessingImage( destFile, srcLine, srcLineIndex,
								d_physicalPageNum, d_psProps,
								(*currentEscapeType), lineY, newLineY );

					if( excessSrcLine ) return( excessSrcLine );
				}

				notCurrentlyEscaped = False;
			}
			else srcLineIndex++;

			break;

		case FORMFEED:
			if( ! PROCESSING_IMAGE(*currentEscapeType) )
			{
				EndLine( destFile, notCurrentlyEscaped, lineX, lineY, d_physicalPageNum, d_psProps );

				(*newLineX) = BODY_LEFT_THRESHOLD;
				(*newLineY) = BOTTOM_THRESHOLD - 1;

				return( NULL );
			}
			else
			{
				if_putc( srcLine[srcLineIndex], destFile );
				srcLineIndex++;
			}

			break;

		case TAB:
			if( ! PROCESSING_IMAGE(*currentEscapeType) )
			{
				while( (currentColumn % d_psProps->tabWidth) != 0 )
				{
					if_putc( ' ', destFile );
					currentLineWidth +=  PSCHARWIDTH( ' ' );
					currentColumn++;
				}

				if_putc( ' ', destFile );

				currentLineWidth += PSCHARWIDTH( ' ' );
				currentColumn++;
			}
			else if_putc( srcLine[srcLineIndex], destFile );

			srcLineIndex++;

			break;

		case '(':
		case ')':
		case '\\':
			if( ! PROCESSING_IMAGE(*currentEscapeType) )
			{
				if_putc( '\\', destFile );
				if_putc( srcLine[srcLineIndex], destFile );

				currentLineWidth += PSCHARWIDTH( srcLine[srcLineIndex] );
				currentColumn++;
			}
			else if_putc( srcLine[srcLineIndex], destFile );

			srcLineIndex++;

			break;

		default:
			if( ! PROCESSING_IMAGE(*currentEscapeType) )
			{
				if_putc( srcLine[srcLineIndex], destFile );

				currentLineWidth += PSCHARWIDTH( srcLine[srcLineIndex] );
				currentColumn++;
			}
			else if_putc( srcLine[srcLineIndex], destFile );

			srcLineIndex++;

			break;
		}
	}

        /*********************************************************************************
         *
         * We finished processing all the characters in the line, so we have to do
	 * some clean up...  If the text is not in NORMAL mode, close off the line and
	 * start a NORMAL line, then immediately end that line.
         *
         *********************************************************************************/

	if( ! PROCESSING_IMAGE(*currentEscapeType) )
	{
		if( (*currentEscapeType) != TEXT_NORMAL )
			EndEscapedStartNormalText( (*currentEscapeType), destFile, d_psProps, d_physicalPageNum );

		EndLine( destFile, notCurrentlyEscaped, lineX, lineY, d_physicalPageNum, d_psProps );

		/* Update the X and Y coordinates for the next input line. */
		(*newLineX) = BODY_LEFT_THRESHOLD;
		(*newLineY) = lineY - BODY_LINE_HEIGHT;
	}
	else if_putc( NEWLINE, destFile );

	if( ferror(destFile) ) g_errorType = XnPS_WRITE_ERROR;

	return( NULL );
}

/***********************************************************/

/*
 *  Proc:	NeedHeaderArea
 *  Callers:	XnGeneratePostScript
 *  Purpose:	Determine if the current physical page needs a header
 *		or not based on the settings of the PostScript properties.
 *  Params:	d_psProps - The PostScript properties.
 *		d_physicalPageNum - The physical page we are testing.
 *  Globals:	g_startingPageNum
 *  Returns:	True if the page needs a header.
 */
Boolean NeedHeaderArea( XnPSPropsStruct *d_psProps, int d_physicalPageNum )
{
	/* We need a header for the page numbers. */
	if( d_psProps->printPageNumbers ) return( True );

	/* If we are not on the first page. */
	if( d_physicalPageNum > g_startingPageNum )
	{
		if( (d_psProps->printTime || d_psProps->printTitle || d_psProps->printUserID) && (d_psProps->allPages) )
			return( True );
		else 
			return( False );
	}

	if( d_psProps->printTime || d_psProps->printTitle || d_psProps->printUserID )
		return( True );

	return( False );
}

/***********************************************************/

/*
 *  Proc:	XnGeneratePostScript
 *  Callers:	<Library>
 *  Purpose:	Convert and ASCII input file into a PostScript output file.
 *  Params:	srcFile - The ASCII input file.
 *		destFile - The PostScript output file.
 *		d_psProps - The PostScript properties structure.
 *		psStatistics - <Return> The PostScript output file statistics.
 *  Globals:	g_DefaultWidths, g_errorType
 *		g_startingPageNum, g_showPageCnt
 *  Returns:	An error code (defined in header) or 0 on success.
 */
int XnGeneratePostScript( FILE *srcFile, FILE *destFile, XnPSPropsStruct *d_psProps, XnPSOutputPropsStruct *psStatistics )
{
	int	lineX;			/* The current x position to place text. */
	int	lineY;			/* The current y position to place text. */
	int	numLinesWrapped;	/* The number of lines wrapped to the next line in the output. */
	int	charsProcessed;		/* The number of characters in the input file. */
	int	linesProcessed;		/* The number of lines output to the PostScript file. */
	int	actualBodyTextWidth;	/* The width of the virtual page (how wide the text was printed). */
	int	virtualPagesProcessed;	/* The number of virtual pages created (can be > 1 per physical page). */
	int	srcLineChars;		/* The number of characters in the current input line. */
	int	currentLineNumber;	/* The line number of the current input line. */
	int	virtualPageNum;		/* What virtual page we are processing. */
	int	d_physicalPageNum;	/* What physical page we are processing. */

	char	*excessSrcLine;
	char	*srcInputLine;		/* The srcFile input line from the file being processed. */
	char	buffer[ MAX_LINE_LEN ];

	float	longestLineWidth;	/* The length of the longest line processed. */

	/*
	 * Describes the current state of the output (Normal, Bold, Underline) mode.
	 * The modes are controled by escape sequences in the input file.
	 * Domain values are TEXT_BOLD, TEXT_ITALIC, TEXT_UNDERLINE, TEXT_NORMAL.
	 */
	int	currentEscapeType;

	/*********************************************************************************
	 *
	 * Make sure our arguments are legal.
	 * 
	 *********************************************************************************/

	if( srcFile == NULL ) return( XnPS_ILLEGAL_SRC );
	if( destFile == NULL ) return( XnPS_ILLEGAL_DEST );
	if( d_psProps == NULL ) return( XnPS_ILLEGAL_PROPS );
	if( psStatistics == NULL ) return( XnPS_ILLEGAL_PROPS );

	/*********************************************************************************
	 *
	 * Initialize all the global variables and statistics values.
	 * 
	 *********************************************************************************/

	g_errorType = 0;
	currentEscapeType = TEXT_NORMAL;

	g_imageCnt = 0;
	g_showPageCnt = 0;
	numLinesWrapped = 0;
	linesProcessed = 0;
	charsProcessed = 0;
	actualBodyTextWidth = 0;
	virtualPagesProcessed = 0;

	longestLineWidth = 0.0;

	if( d_psProps->printPageRange )
		g_startingPageNum = d_psProps->printRangeFrom;
	else g_startingPageNum = 1;

	/*********************************************************************************
	 *
	 * Set up some initial variables.
	 * 
	 *********************************************************************************/

	/* If the user dosen't want character wrapping, set the widths to negative numbers. */
	if( d_psProps->charWidths == NULL ) d_psProps->charWidths = (g_DefaultWidths);
	if( d_psProps->characterWrap == False ) d_psProps->charWidths = (g_DefaultWidths);

	/* Set the width spacing for the line numbers. */
	if( d_psProps->printLineNumbers )
	{
		sprintf( buffer, "%d", (int)d_psProps->maxLineNumber );
		d_psProps->bodyLeftOffsetPts = BodyTextWidthPts( buffer, d_psProps ) + (NUMBERING_LINE_OFFSET * 3);
	}

	/* Start the page numbers at 1. */
	virtualPageNum = 1;
	d_physicalPageNum = 1;

	/* Set the initial x position (where the text will start). */
	lineX = BODY_LEFT_THRESHOLD;

	/* Find out if this page needs a header area. */
	d_psProps->needHeaderArea = NeedHeaderArea( d_psProps, d_physicalPageNum );

	/* Set the initial y position according to the header info. */
	if( d_psProps->needHeaderArea )
		lineY = TOP_THRESHOLD - (HEADER_OFFSET + d_psProps->bodyFontSizePts);
	else
		lineY = TOP_THRESHOLD - d_psProps->bodyFontSizePts;

	/*********************************************************************************
	 *
	 * Generate the PostScript file...
	 * 
	 *********************************************************************************/

	/* Print the comments and procedure definitions to the output file. */
	OutputHeaderPS( destFile, d_psProps );

	/* Output the code to start a page. */
	BeginNewPagePS( destFile, virtualPageNum, d_physicalPageNum, d_psProps );

	fprintf( destFile, "    %d %d moveto  %% Make sure there is a current point.\n", lineX, lineY );

	/* Get the first line of input. */
	srcInputLine = GetInputLine( srcFile, &srcLineChars, &linesProcessed );
	currentLineNumber = linesProcessed;

	while( srcInputLine != NULL )
	{
		/* If there is any type of error during processing, just quit. */
		if( g_errorType )
		{
			XnFreeString( srcInputLine );
			return( g_errorType );
		}

		/* Keep track of how many characters we process. */
		charsProcessed += srcLineChars;

		/*
		 * Process the input line from the srcFile file.  This function returns
		 * three pieces of information:
		 *
		 *	excessSrcLine	- Any text that still needs to be processed.
		 *	lineX	- A new x position for the text.
		 *	lineY	- A new y position for the text.
		 *	numLinesWrapped - The number of lines wrapped (total).
		 *	actualBodyTextWidth - The width of the page for the actual body text.
		 *	longestLineWidth - What the widest line has been (in points).
		 */
		excessSrcLine = ProcessInputLine( destFile, srcInputLine, &currentEscapeType, currentLineNumber,
					d_physicalPageNum, lineX, lineY, d_psProps,
					&lineX, &lineY, &numLinesWrapped, &actualBodyTextWidth,
					&longestLineWidth );

		if( g_errorType ) return( g_errorType );

		/*
		 * If there is a case where a line width is too long to fit on the
		 * page and needs to be wrapped to the next line AND the part of that
		 * line that is being wrapped also runs off the page vertically, it is
		 * returned by ProcessInputLine.  We want to process the rest of that line before getting 
		 * another line from the file.  Update the linesProcessed and the
		 * srcLineChars and charsProcessed (these characters/line were already counted
		 * the first time the line was read).
		 */
		if( excessSrcLine )
		{
			if( !PROCESSING_IMAGE(currentEscapeType) )
			{
				sprintf( buffer, "%s%s", LEADING_SPACES, excessSrcLine );

				XnFreeString( excessSrcLine );
				excessSrcLine = NULL;

				XnFreeString( srcInputLine );
				srcInputLine = XtNewString( buffer );
			}

			srcLineChars = strlen( srcInputLine );
			charsProcessed -= srcLineChars;
			currentLineNumber = (-1);
		}
		else 
		{
			XnFreeString( srcInputLine );
			srcInputLine = GetInputLine( srcFile, &srcLineChars, &linesProcessed );
			currentLineNumber = linesProcessed;
		}

		/*
		 * Check to see if we need to start a new page.  If so, end the current
		 * page and start over.
		 */
		if( lineY < BOTTOM_THRESHOLD )
		{
			if( (feof(srcFile) == 0) || (srcInputLine != NULL) )
			{
				EndPage( destFile, virtualPageNum, d_physicalPageNum, d_psProps, &virtualPageNum, &d_physicalPageNum );

				d_psProps->needHeaderArea = NeedHeaderArea( d_psProps, d_physicalPageNum );

				if( d_psProps->needHeaderArea )
					lineY = TOP_THRESHOLD - (HEADER_OFFSET + d_psProps->bodyFontSizePts);
				else
					lineY = TOP_THRESHOLD - d_psProps->bodyFontSizePts;

				BeginNewPagePS( destFile, virtualPageNum, d_physicalPageNum, d_psProps );

				fprintf( destFile, "    %d %d moveto  %% Make sure there is a current point.\n", lineX, lineY );
			}
		}
	}

	XnFreeString( srcInputLine );
	XnFreeString( excessSrcLine );

	/*
	 * EndPage increments the d_physicalPageNum variable if the next sheet
	 * printed will cause a new sheet to be used.  Since there are
	 * not going to be any more pages, we decrement d_physicalPageNum to offset
	 * the increment in EndPage.  EndPage always increments the sheet
	 * number, but there will be no more sheets, so we must decrement
	 * virtualPageNum to offset this also.
	 */
	EndPage( destFile, virtualPageNum, d_physicalPageNum, d_psProps, &virtualPageNum, &d_physicalPageNum );
	PrintTrailer( destFile, d_physicalPageNum, virtualPageNum - 1, d_physicalPageNum - 1, d_psProps, &d_physicalPageNum );

	/*********************************************************************************
	 *
	 * Fill the statistics structure to return to the caller.
	 * 
	 *********************************************************************************/

	virtualPagesProcessed = virtualPageNum - 1;

	psStatistics->imageCnt = g_imageCnt;
	psStatistics->showPageCnt = g_showPageCnt;
	psStatistics->linesWrapped = numLinesWrapped;
	psStatistics->linesProcessed = linesProcessed;
	psStatistics->charsProcessed = charsProcessed;
	psStatistics->physicalPagesProcessed = g_showPageCnt;
	psStatistics->virtualPagesProcessed = virtualPagesProcessed;
	psStatistics->longestLineWidth = longestLineWidth;
	psStatistics->actualBodyTextWidth = actualBodyTextWidth;

	return( g_errorType );
}

/***********************************************************/

/********
int main(int argc, char *argv[])
{
  FILE			*srcFile;
  FILE			*destFile;
  XnPSPropsStruct	*props;
  XnPSOutputPropsStruct	outputProps;

  if( argc != 3 ) exit( -1 );

  srcFile=fopen(argv[1], "r");
  destFile=fopen(argv[2], "w");
  if( srcFile == NULL ) exit( -2 );
  if( destFile == NULL ) exit( -3 );

  props=XnCreatePSProps( XnPS_TWO_ONE );

  XnGeneratePostScript(srcFile, destFile, props, &outputProps);

  fclose( srcFile );
  fclose( destFile );

  printf("%.2f longest line width.\n", outputProps.longestLineWidth);
  printf("%d actual body text width\n", outputProps.actualBodyTextWidth);
  printf("%d lines processed.  %d chars processed.  %d lines wrapped.\n",
	 outputProps.linesProcessed, outputProps.charsProcessed, outputProps.linesWrapped);
  printf("%d virtual pages processed.  %d physical pages processed.\n",
	outputProps.virtualPagesProcessed, outputProps.physicalPagesProcessed);
}
***********/

