/*
 *  NXPostScript - nxpsutils.c 
 *
 *	Version 2.0
 *      by Lance Lovette
 *	Code and Display Copyright (C) 1993 by Lance Lovette
 *	All Rights Reserved
 *
 *	Lance Lovette (lovette@iftech.com)
 *	North Carolina State University
 *	Engineering Computer Operations
 *	Department of Computer Science
 *	Raleigh, NC 27607
 *
 *  Permission to use, copy, modify and distribute this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both the
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the names of Lance Lovette and North Carolina
 *  State University not be used in advertising or publicity pertaining to
 *  distribution of the software without specific, written prior permission.
 *  Lance Lovette and North Carolina State University make no
 *  representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 *  Lance Lovette AND North Carolina State University DISCLAIMS ALL
 *  WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Lance Lovette OR
 *  North Carolina State University BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 *  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 *  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 *  OF THIS SOFTWARE.
 *
 *  $RCSfile: nxpsutils.c,v $
 *  $Revision: 2.2 $
 *  $Date: 1994/01/02 20:01:59 $
 */
static char rcsid[] = "$Id: nxpsutils.c,v 2.2 1994/01/02 20:01:59 ltlovett Exp $";

#include "NXUtils.h"
#include "NXPostScript.h"
#include "nxpsutils.h"

#include <ctype.h>

/***********************************************************/

/*
 * The following arrays contain the widths of characters in that typeface.
 * The widths are of a character that is 1 point tall (Eg. the width of a 10 point
 * font character is 10 * array[x]).  The arrays are indexed by the ordinal
 * value of the character (Eg. 'A'=65).  The arrays have a length of 256.  The values in the
 * arrays were taken from the Adobe font metric files in the directory:
 *	/usr/local/lib/transcript/ *.afm
 * These widths are used in character wrap mode to calculate the length of the
 * current line.  The defaultWidths are used when not in character wrap mode.
 */

float g_DefaultWidths[]={
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1};

float g_CourierWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,
	0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,
	0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,
	0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,
	0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,
	0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.000,0.000,0.000,0.600,0.600,0.600,
	0.600,0.000,0.600,0.600,0.600,0.600,0.600,0.600,0.600,0.000,0.000,0.600,0.000,0.600,0.600,0.600,0.600,0.600,0.600,0.600,
	0.600,0.000,0.600,0.600,0.000,0.600,0.600,0.600,0.600,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.600,0.000,0.000,0.000,0.000,0.600,0.600,0.000,0.600,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.600,0.000,0.000,0.600,0.600,0.000,0.600,0.000,0.000,0.000,0.000};

float g_HelveticaWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.278,0.278,0.355,0.556,0.556,0.889,0.667,0.222,
	0.333,0.333,0.389,0.584,0.278,0.333,0.278,0.278,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.278,0.278,
	0.584,0.584,0.584,0.556,1.015,0.667,0.667,0.722,0.722,0.667,0.611,0.778,0.722,0.278,0.500,0.667,0.556,0.833,0.722,0.778,
	0.667,0.778,0.722,0.667,0.611,0.722,0.667,0.944,0.667,0.667,0.611,0.278,0.278,0.278,0.469,0.556,0.222,0.556,0.556,0.500,
	0.556,0.556,0.278,0.556,0.556,0.222,0.222,0.500,0.222,0.833,0.556,0.556,0.556,0.556,0.333,0.500,0.278,0.556,0.500,0.722,
	0.500,0.500,0.500,0.334,0.260,0.334,0.584,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.333,0.556,0.556,0.167,0.556,0.556,0.556,0.556,0.191,0.333,0.556,0.333,0.333,0.500,0.500,0.000,0.556,0.556,0.556,
	0.278,0.000,0.537,0.350,0.222,0.333,0.333,0.556,1.000,1.000,0.000,0.611,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,1.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,1.000,0.000,0.370,0.000,0.000,0.000,0.000,0.556,0.778,1.000,0.365,0.000,0.000,0.000,0.000,
	0.000,0.889,0.000,0.000,0.000,0.278,0.000,0.000,0.222,0.611,0.944,0.611,0.000,0.000,0.000,0.000};

float g_HelveticaBoldWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.278,0.333,0.474,0.556,0.556,0.889,0.722,0.278,
	0.333,0.333,0.389,0.584,0.278,0.333,0.278,0.278,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.333,0.333,
	0.584,0.584,0.584,0.611,0.975,0.722,0.722,0.722,0.722,0.667,0.611,0.778,0.722,0.278,0.556,0.722,0.611,0.833,0.722,0.778,
	0.667,0.778,0.722,0.667,0.611,0.722,0.667,0.944,0.667,0.667,0.611,0.333,0.278,0.333,0.584,0.556,0.278,0.556,0.611,0.556,
	0.611,0.556,0.333,0.611,0.611,0.278,0.278,0.556,0.278,0.889,0.611,0.611,0.611,0.611,0.389,0.556,0.333,0.611,0.556,0.778,
	0.556,0.556,0.500,0.389,0.280,0.389,0.584,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.333,0.556,0.556,0.167,0.556,0.556,0.556,0.556,0.238,0.500,0.556,0.333,0.333,0.611,0.611,0.000,0.556,0.556,0.556,
	0.278,0.000,0.556,0.350,0.278,0.500,0.500,0.556,1.000,1.000,0.000,0.611,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,1.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,1.000,0.000,0.370,0.000,0.000,0.000,0.000,0.611,0.778,1.000,0.365,0.000,0.000,0.000,0.000,
	0.000,0.889,0.000,0.000,0.000,0.278,0.000,0.000,0.278,0.611,0.944,0.611,0.000,0.000,0.000,0.000};

float g_HelveticaBoldObliqueWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.278,0.333,0.474,0.556,0.556,0.889,0.722,0.278,
	0.333,0.333,0.389,0.584,0.278,0.333,0.278,0.278,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.333,0.333,
	0.584,0.584,0.584,0.611,0.975,0.722,0.722,0.722,0.722,0.667,0.611,0.778,0.722,0.278,0.556,0.722,0.611,0.833,0.722,0.778,
	0.667,0.778,0.722,0.667,0.611,0.722,0.667,0.944,0.667,0.667,0.611,0.333,0.278,0.333,0.584,0.556,0.278,0.556,0.611,0.556,
	0.611,0.556,0.333,0.611,0.611,0.278,0.278,0.556,0.278,0.889,0.611,0.611,0.611,0.611,0.389,0.556,0.333,0.611,0.556,0.778,
	0.556,0.556,0.500,0.389,0.280,0.389,0.584,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.333,0.556,0.556,0.167,0.556,0.556,0.556,0.556,0.238,0.500,0.556,0.333,0.333,0.611,0.611,0.000,0.556,0.556,0.556,
	0.278,0.000,0.556,0.350,0.278,0.500,0.500,0.556,1.000,1.000,0.000,0.611,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,1.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,1.000,0.000,0.370,0.000,0.000,0.000,0.000,0.611,0.778,1.000,0.365,0.000,0.000,0.000,0.000,
	0.000,0.889,0.000,0.000,0.000,0.278,0.000,0.000,0.278,0.611,0.944,0.611,0.000,0.000,0.000,0.000};

float g_HelveticaObliqueWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.278,0.278,0.355,0.556,0.556,0.889,0.667,0.222,
	0.333,0.333,0.389,0.584,0.278,0.333,0.278,0.278,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.556,0.278,0.278,
	0.584,0.584,0.584,0.556,1.015,0.667,0.667,0.722,0.722,0.667,0.611,0.778,0.722,0.278,0.500,0.667,0.556,0.833,0.722,0.778,
	0.667,0.778,0.722,0.667,0.611,0.722,0.667,0.944,0.667,0.667,0.611,0.278,0.278,0.278,0.469,0.556,0.222,0.556,0.556,0.500,
	0.556,0.556,0.278,0.556,0.556,0.222,0.222,0.500,0.222,0.833,0.556,0.556,0.556,0.556,0.333,0.500,0.278,0.556,0.500,0.722,
	0.500,0.500,0.500,0.334,0.260,0.334,0.584,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.333,0.556,0.556,0.167,0.556,0.556,0.556,0.556,0.191,0.333,0.556,0.333,0.333,0.500,0.500,0.000,0.556,0.556,0.556,
	0.278,0.000,0.537,0.350,0.222,0.333,0.333,0.556,1.000,1.000,0.000,0.611,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,1.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,1.000,0.000,0.370,0.000,0.000,0.000,0.000,0.556,0.778,1.000,0.365,0.000,0.000,0.000,0.000,
	0.000,0.889,0.000,0.000,0.000,0.278,0.000,0.000,0.222,0.611,0.944,0.611,0.000,0.000,0.000,0.000};

float g_TimesRomanWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.250,0.333,0.408,0.500,0.500,0.833,0.778,0.333,
	0.333,0.333,0.500,0.564,0.250,0.333,0.250,0.278,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.278,0.278,
	0.564,0.564,0.564,0.444,0.921,0.722,0.667,0.667,0.722,0.611,0.556,0.722,0.722,0.333,0.389,0.722,0.611,0.889,0.722,0.722,
	0.556,0.722,0.667,0.556,0.611,0.722,0.722,0.944,0.722,0.722,0.611,0.333,0.278,0.333,0.469,0.500,0.333,0.444,0.500,0.444,
	0.500,0.444,0.333,0.500,0.500,0.278,0.278,0.500,0.278,0.778,0.500,0.500,0.500,0.500,0.333,0.389,0.278,0.500,0.500,0.722,
	0.500,0.500,0.444,0.480,0.200,0.480,0.541,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.333,0.500,0.500,0.167,0.500,0.500,0.500,0.500,0.180,0.444,0.500,0.333,0.333,0.556,0.556,0.000,0.500,0.500,0.500,
	0.250,0.000,0.453,0.350,0.333,0.444,0.444,0.500,1.000,1.000,0.000,0.444,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,1.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.889,0.000,0.276,0.000,0.000,0.000,0.000,0.611,0.722,0.889,0.310,0.000,0.000,0.000,0.000,
	0.000,0.667,0.000,0.000,0.000,0.278,0.000,0.000,0.278,0.500,0.722,0.500,0.000,0.000,0.000,0.000};

float g_TimesBoldWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.250,0.333,0.555,0.500,0.500,1.000,0.833,0.333,
	0.333,0.333,0.500,0.570,0.250,0.333,0.250,0.278,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.333,0.333,
	0.570,0.570,0.570,0.500,0.930,0.722,0.667,0.722,0.722,0.667,0.611,0.778,0.778,0.389,0.500,0.778,0.667,0.944,0.722,0.778,
	0.611,0.778,0.722,0.556,0.667,0.722,0.722,1.000,0.722,0.722,0.667,0.333,0.278,0.333,0.581,0.500,0.333,0.500,0.556,0.444,
	0.556,0.444,0.333,0.500,0.556,0.278,0.333,0.556,0.278,0.833,0.556,0.500,0.556,0.556,0.444,0.389,0.333,0.556,0.500,0.722,
	0.500,0.500,0.444,0.394,0.220,0.394,0.520,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.333,0.500,0.500,0.167,0.500,0.500,0.500,0.500,0.278,0.500,0.500,0.333,0.333,0.556,0.556,0.000,0.500,0.500,0.500,
	0.250,0.000,0.540,0.350,0.333,0.500,0.500,0.500,1.000,1.000,0.000,0.500,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,1.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,1.000,0.000,0.300,0.000,0.000,0.000,0.000,0.667,0.778,1.000,0.330,0.000,0.000,0.000,0.000,
	0.000,0.722,0.000,0.000,0.000,0.278,0.000,0.000,0.278,0.500,0.722,0.556,0.000,0.000,0.000,0.000};

float g_TimesItalicWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.250,0.333,0.420,0.500,0.500,0.833,0.778,0.333,
	0.333,0.333,0.500,0.675,0.250,0.333,0.250,0.278,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.333,0.333,
	0.675,0.675,0.675,0.500,0.920,0.611,0.611,0.667,0.722,0.611,0.611,0.722,0.722,0.333,0.444,0.667,0.556,0.833,0.667,0.722,
	0.611,0.722,0.611,0.500,0.556,0.722,0.611,0.833,0.611,0.556,0.556,0.389,0.278,0.389,0.422,0.500,0.333,0.500,0.500,0.444,
	0.500,0.444,0.278,0.500,0.500,0.278,0.278,0.444,0.278,0.722,0.500,0.500,0.500,0.500,0.389,0.389,0.278,0.500,0.444,0.667,
	0.444,0.444,0.389,0.400,0.275,0.400,0.541,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.389,0.500,0.500,0.167,0.500,0.500,0.500,0.500,0.214,0.556,0.500,0.333,0.333,0.500,0.500,0.000,0.500,0.500,0.500,
	0.250,0.000,0.523,0.350,0.333,0.556,0.556,0.500,0.889,1.000,0.000,0.500,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,0.889,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.889,0.000,0.276,0.000,0.000,0.000,0.000,0.556,0.722,0.944,0.310,0.000,0.000,0.000,0.000,
	0.000,0.667,0.000,0.000,0.000,0.278,0.000,0.000,0.278,0.500,0.667,0.500,0.000,0.000,0.000,0.000};

float g_TimesBoldItalicWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.250,0.389,0.555,0.500,0.500,0.833,0.778,0.333,
	0.333,0.333,0.500,0.570,0.250,0.333,0.250,0.278,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.333,0.333,
	0.570,0.570,0.570,0.500,0.832,0.667,0.667,0.667,0.722,0.667,0.667,0.722,0.778,0.389,0.500,0.667,0.611,0.889,0.722,0.722,
	0.611,0.722,0.667,0.556,0.611,0.722,0.667,0.889,0.667,0.611,0.611,0.333,0.278,0.333,0.570,0.500,0.333,0.500,0.500,0.444,
	0.500,0.444,0.333,0.500,0.556,0.278,0.278,0.500,0.278,0.778,0.556,0.500,0.500,0.500,0.389,0.389,0.278,0.556,0.444,0.667,
	0.500,0.444,0.389,0.348,0.220,0.348,0.570,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.389,0.500,0.500,0.167,0.500,0.500,0.500,0.500,0.278,0.500,0.500,0.333,0.333,0.556,0.556,0.000,0.500,0.500,0.500,
	0.250,0.000,0.500,0.350,0.333,0.500,0.500,0.500,1.000,1.000,0.000,0.500,0.000,0.333,0.333,0.333,0.333,0.333,0.333,0.333,
	0.333,0.000,0.333,0.333,0.000,0.333,0.333,0.333,1.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.944,0.000,0.266,0.000,0.000,0.000,0.000,0.611,0.722,0.944,0.300,0.000,0.000,0.000,0.000,
	0.000,0.722,0.000,0.000,0.000,0.278,0.000,0.000,0.278,0.500,0.722,0.500,0.000,0.000,0.000,0.000};

float g_SymbolWidths[]={
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.250,0.333,0.713,0.500,0.549,0.833,0.778,0.439,
	0.333,0.333,0.500,0.549,0.250,0.549,0.250,0.278,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.500,0.278,0.278,
	0.549,0.549,0.549,0.444,0.549,0.696,0.660,0.710,0.612,0.652,0.763,0.603,0.765,0.351,0.631,0.724,0.686,0.918,0.739,0.750,
	0.768,0.741,0.580,0.592,0.632,0.690,0.439,0.768,0.645,0.795,0.650,0.333,0.863,0.333,0.658,0.500,0.500,0.631,0.549,0.549,
	0.494,0.439,0.521,0.411,0.603,0.329,0.603,0.549,0.549,0.576,0.521,0.549,0.549,0.521,0.549,0.603,0.439,0.576,0.713,0.686,
	0.493,0.686,0.494,0.480,0.200,0.480,0.549,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,0.000,
	0.000,0.620,0.247,0.549,0.167,0.713,0.500,0.753,0.753,0.753,0.753,1.042,0.987,0.603,0.987,0.603,0.400,0.549,0.411,0.549,
	0.549,0.713,0.494,0.460,0.549,0.549,0.549,0.549,1.000,0.603,1.000,0.658,0.823,0.686,0.795,0.987,0.768,0.768,0.823,0.768,
	0.768,0.713,0.713,0.713,0.713,0.713,0.713,0.713,0.768,0.713,0.790,0.790,0.890,0.823,0.549,0.250,0.713,0.603,0.603,1.042,
	0.987,0.603,0.987,0.603,0.494,0.329,0.790,0.790,0.786,0.713,0.384,0.384,0.384,0.384,0.384,0.384,0.494,0.494,0.494,0.494,
	0.790,0.329,0.274,0.686,0.686,0.686,0.384,0.384,0.384,0.384,0.384,0.384,0.494,0.494,0.494,0.000};

/***********************************************************/

/*
 *  Proc:	XnPSFontCharWidths
 *  Callers:	<Library>
 *  Purpose:	Returns an array that specifies the character widths for a
 *		particular font style.
 *  Params:	fontname - The name of the font to return.
 *  Globals:	g_* - The arrays containing the widths.
 *  Returns:	The corresponding array or NULL if one is not found.
 */
float *XnPSFontCharWidths( char *fontname )
{
	if( !fontname ) return( (float *)NULL );

	if( strcmp(fontname, "Courier"			) == 0 ) return( g_CourierWidths ); else
	if( strcmp(fontname, "Courier-Bold"		) == 0 ) return( g_CourierWidths ); else
	if( strcmp(fontname, "Courier-Oblique"		) == 0 ) return( g_CourierWidths ); else
	if( strcmp(fontname, "Courier-BoldOblique"	) == 0 ) return( g_CourierWidths ); else

	if( strcmp(fontname, "Helvetica"		) == 0 ) return( g_HelveticaWidths ); else
	if( strcmp(fontname, "Helvetica-Bold"		) == 0 ) return( g_HelveticaBoldWidths ); else
	if( strcmp(fontname, "Helvetica-Oblique"	) == 0 ) return( g_HelveticaObliqueWidths ); else
	if( strcmp(fontname, "Helvetica-BoldOblique"	) == 0 ) return( g_HelveticaBoldObliqueWidths ); else

	if( strcmp(fontname, "Times-Roman"		) == 0 ) return( g_TimesRomanWidths ); else
	if( strcmp(fontname, "Times-Bold"		) == 0 ) return( g_TimesBoldWidths ); else
	if( strcmp(fontname, "Times-Italic"		) == 0 ) return( g_TimesItalicWidths ); else
	if( strcmp(fontname, "Times-BoldItalic"		) == 0 ) return( g_TimesBoldItalicWidths ); else

	if( strcmp(fontname, "Symbol"			) == 0 ) return( g_SymbolWidths );

	return( (float *)NULL );
}

/***********************************************************/

#define	S_NORMAL		0
#define	S_BOLD			1
#define	S_ITALIC		2
#define	S_BOLDITALIC		3

#define	BOLD			"Bold"
#define	OBLIQUE			"Oblique"
#define	DEMI			"Demi"
#define	BOOK			"Book"
#define	ITALIC			"Italic"
#define	ROMAN			"Roman"
#define	BOLDITALIC		"BoldItalic"
#define	BOLDOBLIQUE		"BoldOblique"
#define BOOKOBLIQUE		"BookOblique"
#define	DEMIOBLIQUE		"DemiOblique"
#define	DEMIITALIC		"DemiItalic"
#define LIGHT			"Light"
#define LIGHTITALIC		"LightItalic"

/*
 *  Proc:	XnCreatePSFontname
 *  Callers:	<Library>
 *  Purpose:	Create a PostScript font name that corresponds to 
 *		a particular font name and style.  For example
 *		the font name "Times" and style of "Bold" would
 *		be "TimesBold" in PostScript world.
 *  Params:	orig_fontname - The font name.
 *		orig_fontstyle - The font style.
 *  Globals:	None.
 *  Returns:	The PostScript font name or NULL if one could not
 *		be created that matches the requested name/style.
 */
char *XnCreatePSFontname( char *orig_fontname, char *orig_fontstyle )
{
	int	i;
	int	styletype	=0;
	char	*newfont	=NULL;
	char	*fontname	=NULL;
	char	*fontstyle	=NULL;

	if( !orig_fontname || !orig_fontstyle ) return( NULL );

	fontname = XtNewString( orig_fontname );
	fontstyle = XtNewString( orig_fontstyle );

	/* Change the font name to all uppercase. */
	for( i=0; fontname[i]; i++ )
	{
		if( fontname[i] == ' ' ) fontname[i] = '\0';
		else fontname[i] = toupper( fontname[i] );
	}

	/* Change the font style to all uppercase. */
	for( i=0; fontstyle[i]; i++ )
	{
		if( fontstyle[i] == ' ' ) fontstyle[i] = '\0';
		else fontstyle[i] = toupper( fontstyle[i] );
	}

	if( strcmp( "NORMAL", fontstyle		) == 0 ) styletype = S_NORMAL;		else
	if( strcmp( "BOLD", fontstyle		) == 0 ) styletype = S_BOLD;		else
	if( strcmp( "ITALIC", fontstyle		) == 0 ) styletype = S_ITALIC;		else
	if( strcmp( "BOLDITALIC", fontstyle	) == 0 ) styletype = S_BOLDITALIC;
	else return( NULL );

	newfont = XnMallocString( strlen(fontname) + 25 ); 

	if( strcmp( "TIMES", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "Times-%s", ROMAN );	else
		if( styletype == S_BOLD		) sprintf( newfont, "Times-%s", BOLD );		else
		if( styletype == S_ITALIC	) sprintf( newfont, "Times-%s", ITALIC );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "Times-%s", BOLDITALIC );
		else return( NULL );
	}

	else
	if( strcmp( "HELVETICA", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "Helvetica" );		else
		if( styletype == S_BOLD		) sprintf( newfont, "Helvetica-%s", BOLD );	else 
		if( styletype == S_ITALIC	) sprintf( newfont, "Helvetica-%s", OBLIQUE );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "Helvetica-%s", BOLDOBLIQUE );
		else return( NULL );
	}

	else
	if( strcmp( "COURIER", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "Courier" );		else
		if( styletype == S_BOLD		) sprintf( newfont, "Courier-%s",  BOLD );	else 
		if( styletype == S_ITALIC	) sprintf( newfont, "Courier-%s",  OBLIQUE );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "Courier-%s",  BOLDOBLIQUE );
		else return( NULL );
	}

	else
	if( strcmp( "AVANTGARDE", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "AvantGarde-%s", BOOK );		else
		if( styletype == S_BOLD		) sprintf( newfont, "AvantGarde-%s", DEMI );		else 
		if( styletype == S_ITALIC	) sprintf( newfont, "AvantGarde-%s", BOOKOBLIQUE );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "AvantGarde-%s", DEMIOBLIQUE );
		else return( NULL );
	}

	else
	if( strcmp( "PALATINO", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "Palatino-%s", ROMAN );	else
		if( styletype == S_BOLD		) sprintf( newfont, "Palatino-%s", BOLD );	else 
		if( styletype == S_ITALIC	) sprintf( newfont, "Palatino-%s", ITALIC );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "Palatino-%s", BOLDITALIC );
		else return( NULL );
	}

	else
	if( strcmp( "NEWCENTURYSCHLBK", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "NewCenturySchlbk-%s", ROMAN );	else
		if( styletype == S_BOLD		) sprintf( newfont, "NewCenturySchlbk-%s", BOLD );	else 
		if( styletype == S_ITALIC	) sprintf( newfont, "NewCenturySchlbk-%s", ITALIC );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "NewCenturySchlbk-%s", BOLDITALIC );
		else return( NULL );
	}

	else
	if( strcmp( "LUBALINGRAPH", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "LubalinGraph-%s", BOOK );		else
		if( styletype == S_BOLD		) sprintf( newfont, "LubalinGraph-%s", DEMI );		else 
		if( styletype == S_ITALIC	) sprintf( newfont, "LubalinGraph-%s", BOOKOBLIQUE );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "LubalinGraph-%s", DEMIOBLIQUE );
		else return( NULL );
	}

	else
	if( strcmp( "HELVETICA-NARROW", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "Helvetica-Narrow" );		else
		if( styletype == S_BOLD		) sprintf( newfont, "Helvetica-Narrow-%s",  BOLD );	else 
		if( styletype == S_ITALIC	) sprintf( newfont, "Helvetica-Narrow-%s",  OBLIQUE );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "Helvetica-Narrow-%s",  BOLDOBLIQUE );
		else return( NULL );
	}

	else
	if( strcmp( "SOUVENIR", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "Souvenir-%s", LIGHT );		else
		if( styletype == S_BOLD		) sprintf( newfont, "Souvenir-%s", DEMI );		else 
		if( styletype == S_ITALIC	) sprintf( newfont, "Souvenir-%s", LIGHTITALIC );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "Souvenir-%s", DEMIITALIC );
		else return( NULL );
	}

	else
	if( strcmp( "BOOKMAN", fontname ) == 0 )
	{
		if( styletype == S_NORMAL	) sprintf( newfont, "Bookman-%s", LIGHT );		else
		if( styletype == S_BOLD		) sprintf( newfont, "Bookman-%s", DEMI );		else 
		if( styletype == S_ITALIC	) sprintf( newfont, "Bookman-%s", LIGHTITALIC );	else
		if( styletype == S_BOLDITALIC	) sprintf( newfont, "Bookman-%s", DEMIITALIC );
		else return( NULL );
	}

	else
	if( strcmp( "ZAPFCHANCERY", fontname ) == 0 )
	{
		sprintf( newfont, "ZapfChancery-%s", "MediumItalic" );
	}

	else
	{
		XnFreeString( newfont );
		newfont = NULL;
	}

	XnFreeString( fontname );
	XnFreeString( fontstyle );

	return( newfont );
}

/***********************************************************/

/*
 *  Proc:	XnIsPostScriptFile
 *  Callers:	<Library>
 *  Purpose:	Determine if the specified file is a PostScript file.
 *  Params:	filename - The name of the file.
 *  Globals:	None.
 *  Returns:	True if the file is PostScript.
 */
Boolean XnIsPostScriptFile( char *filename )
{
	char          t[5];
	char		*s	=NULL;
	FILE          *file   =NULL;

	if( !filename ) return( False );

	file = fopen( filename, "r" );
	if( file )
	{
		s = fgets( t, 5, file );
		fclose( file );

		if( s && (strncmp(t, "%!", 2) == 0) )
			return( True );
		else
			return( False );
	}

	return( False );
}

/***********************************************************/

/*
 *  Proc:	XnPagesInPostScriptFile
 *  Callers:	<Library>
 *  Purpose:	Determine the number of pages in a PostScript file.
 *  Params:	filename - The name of the file.
 *  Globals:	None.
 *  Returns:	The number of pages of PostScript or -1 if the
 *		count could not be determined.
 */
int XnPagesInPostScriptFile( char *filename )
{
	int           pages   =(-1);
	char          t[256];
	char		s[256];
	FILE          *file   =NULL;

	if( !filename ) return( -1 );

	if( XnIsPostScriptFile(filename) == False ) return( -1 );

	file = fopen( filename, "r" );
	if( file )
	{
		while( fgets(t, 256, file) )
		{
			if( t[ strlen(t) - 1 ] == '\n' ) t[ strlen(t) - 1 ] = '\0';

			if( (t[6] == 's') && (t[2] == 'P') && (strncmp(t, "%%Pages:", 8) == 0) )
			{
				sscanf( t, "%%%%Pages:%s", s );
				if( strcmp(s, "(atend)") != 0 )
				{
					sscanf( s, "%d", &pages );
					fclose( file );
					return( pages );
				}
			}
		}
		fclose( file );
	}

	return( pages );
}

