/*
 *  NXUtils - NXUtils.h 
 *
 *  $RCSfile: NXUtils.h,v $
 *  $Revision: 2.2 $
 *  $Date: 1994/01/02 20:03:03 $
 */
#ifndef _NXUTILS_H_
#define _NXUTILS_H_

#include <time.h>
#include <pwd.h>
#include <strings.h>
#include <errno.h>

#include <Xm/Xm.h>
#include <X11/StringDefs.h>

/**************************************************************************/

typedef String * StringTable;		/* An array of strings. */

/**************************************************************************/

/*
 * Simple Macros for string manipulation.
 *	XnEndChar - Index of the last character in the string or -1 if there is no string.
 *	XnFreeString - Free's a String.
 *	XnFreeStringTable - Free's an array of n Strings.
 */
#define	XnEndChar( s )			( ((String)(s)) ? (strlen((String)(s)) - 1) : (-1) )
#define	XnFreeString( s )		{ if( (String)(s) ) XtFree( (String)(s) ); }
#define	XnFreeStringTable( s, n )	{ int _i_; for( _i_=0; (s) && (_i_ < (n)); _i_++ ) XnFreeString( (String)((StringTable)s)[_i_] ); }

/*
 * Simple Macros for referencing the fields in various Xm Callback structures.
 *	XnScaleCallbackValue - Returns the 'value' field of an XmScaleCallbackStruct or -1 on error.
 *	XnToggleButtonCallbackSet - Returns the 'set' field of an XmToggleButtonCallbackStruct or False on error.
 *	XnListCallbackItemString - Returns the 'item' field of an XmListCallbackStruct as a string or NULL on error.
 *	XnSelectionBoxCallbackValueString - Returns the 'value' of an XmSelectionBoxCallbackStruct as a string or NULL on error.
 */
#define	XnScaleCallbackValue( d )		( ((XmScaleCallbackStruct *)(d)) ?\
							(((XmScaleCallbackStruct *)(d))->value) : (int)-1 )
#define	XnToggleButtonCallbackSet( d )		( ((XmToggleButtonCallbackStruct *)(d)) ?\
							(((XmToggleButtonCallbackStruct *)(d))->set) : (Boolean)False )
#define	XnListCallbackItemString( d )		( ((XmListCallbackStruct *)d) ? (XnGetXmString(((XmListCallbackStruct *)(d))->item)) : (String)NULL )
#define	XnSelectionBoxCallbackValueString( d )	( ((XmSelectionBoxCallbackStruct *)(d)) ?\
							(XnGetXmString(((XmSelectionBoxCallbackStruct *)(d))->value)) : (String)NULL )

/*
 * Simple Macros for examining the 'reason' field in various Xm Callback structures.
 *	XnDialogCallbackReason - Returns the 'reason' field of an XmAnyCallbackStruct or 0 on error.
 *	XnDialogCallbackReasonHelp - Non-zero if the callback reason is XmCR_HELP.
 *	XnDialogCallbackReasonOK - Non-zero if the callback reason is XmCR_OK.
 *	XnDialogCallbackReasonCancel - Non-zero if the callback reason is XmCR_CANCEL.
 *	XnScaleCallbackReason - Returns the 'reason' field of an XmScaleCallbackStruct or 0 on error.
 */
#define	XnDialogCallbackReason( d )		( ((XmAnyCallbackStruct *)(d)) ? (((XmAnyCallbackStruct *)(d))->reason) : (int)0 )
#define	XnDialogCallbackReasonHelp( d )		( ((XmAnyCallbackStruct *)(d)) ? (((XmAnyCallbackStruct *)(d))->reason == XmCR_HELP) : (int)0 )
#define	XnDialogCallbackReasonOK( d )		( ((XmAnyCallbackStruct *)(d)) ? (((XmAnyCallbackStruct *)(d))->reason == XmCR_OK) : (int)0 )
#define	XnDialogCallbackReasonCancel( d )	( ((XmAnyCallbackStruct *)(d)) ? (((XmAnyCallbackStruct *)(d))->reason == XmCR_CANCEL) : (int)0 )
#define	XnScaleCallbackReason( d )		( ((XmScaleCallbackStruct *)(d)) ? (((XmScaleCallbackStruct *)(d))->reason) : (int)0 )

/*
 * Retrieves the contents of the Motif clipboard that have the specified 'format'.
 * The contents of the clipboard are returned or NULL if there is an error (eg. any of the clipboard
 * retrieval functions do not return ClipboardSuccess).  The length of the returned
 * string is returned in 'return_len'.  If the clipboard is empty, the empty string ("") is returned.
 */
extern String XnRetrieveClipboardContents	(Widget widget, Time timestamp, String format, unsigned long *return_len);


/*
 * Returns the current working directory that the file selection box  is in. 
 */
extern String XnFileSelectionBoxCWD		(Widget fsb);


/* 
 * File manipulation functions:
 *	XnFileSize - Returns the size of a file, or a negative value if an error occurs.
 *	XnFileExists - Returns True if the file 'filename' exists and False if it does not.
 *	XnFileWritable - Returns True if the file 'filename' is writable and False if is does not.
 *			If the file exists, it checks the file permissions.  If it dosen't exist
 *			it checks to see if the file could be opened for writing.
 *	XnFileReadable - Returns True if the file 'filename' is readable (has read permission)
 *			and False if is does not.
 *	XnDeleteFile - Deletes a file/directory.  Returns True if successful.
 *	XnReadFile - Reads a file and returns its contents.  If the file can't be read
 *			or is empty, a default value is returned.  The return value
 *			will be 0 if everything is successful (i.e no read errors).
 *	XnWriteFile - Writes a text string to a file.  The write mode specifies what
 *			mode the file should be opened in for writing (i.e. create, append, ...).
 *			The return value will be 0 on success.  The function returns
 *			the number of items written.
 */
extern int	XnFileSize			( String filename );
extern Boolean	XnFileExists			( String filename );
extern Boolean	XnFileWritable			( String filename );
extern Boolean	XnFileReadable			( String filename );
extern Boolean	XnDeleteFile			( String filename );
extern String 	XnReadFile			( String filename, String defaultValue, int *returnValue );
extern int	XnWriteFile			( String filename, String text, String writemode, int *returnValue );

/*
 * Widget creation functions:
 *	XnMakeLabel - Creates a label
 *	XnMakeToggleButton - Creates a toggle button
 *	XnMakePushButton -  Creates a push button
 */
extern Widget	XnMakeLabel			(Widget parent, String name, String label, Arg *argList, int argCount);
extern Widget	XnMakeToggleButton		(Widget parent, String name, String label, 
							void (*value_changed_callback)(Widget, XtPointer, XmToggleButtonCallbackStruct *),
							XtPointer clientData, Arg *argList, int argCount);
extern Widget	XnMakePushButton		(Widget parent, String name, String label, 
							void (*activate_callback)(Widget, XtPointer, XmPushButtonCallbackStruct *),
							XtPointer clientData, Arg *argList, int argCount);

/*
 * Menu creation functions:
 *	XnMakeMenu - Creates a pulldown menu.  Returns the menu.
 *	XnMakeHelpMenu - Creates a help pulldown menu. Returns the menu.
 *	XnAddAccelerator - Adds an accelerator to 'widget'.
 *	XnMakeMenuOption - Creates a push button option for a menu.  Returns the button.
 */
extern Widget	XnMakeMenu			(Widget parent, String name, String label, KeySym mnemonic);
extern Widget	XnMakeHelpMenu			(Widget parent, String name, String label, KeySym mnemonic);
extern Boolean	XnAddAccelerator		(Widget widget, String acceleratorText, String acceleratorKey);
extern Widget	XnMakeMenuOption		(Widget parent, String name, String label, KeySym mnemonic, 
							void (*activate_callback)(Widget, XtPointer, XmPushButtonCallbackStruct *),
							XtPointer clientData);

/*
 * Widget resource manipulation functions:
 *	XnSetLabelString - Sets the XmNlabelString resource for a widget to 'string'.
 *	XnSetDialogText - Set the XmNtextString resource for a selection box dialog XmDIALOG_TEXT widget
 *				to 'string'.  Returns the associated text widget.
 *	XnGetDialogText - Get the XmNtextString resource for a selection box dialog XmDIALOG_TEXT widget.
 *				Returns the string or NULL if it is empty.
 *	XnSetDialogType - Sets the XmNdialogType of 'widget' to 'type'.
 *	XnSetDialogMessageString - Set the XmNmessageString resource for a dialog widget to 'string'.
 *	XnScaleSetValues - Sets the XmNvalue, XmNminimum, and XmNmaximum for 'scale'.
 */
extern void	XnSetLabelString		( Widget label, String string );
extern void	XnSetDialogMessageString	( Widget dialog, String string );
extern String 	XnGetDialogText			( Widget dialog );
extern void	XnSetDialogType			( Widget dialog, unsigned char type );
extern Widget	XnSetDialogText			( Widget dialog, String string );
extern void	XnScaleSetValues		( Widget scale, int value, int min, int max );

/*
 * String manipulation functions:
 *	XnStrnum - Returns the number of times 'pattern' appears in 'string'.
 *	XnDupString - Duplicates string; Like XtNewString except not a macro.
 *				Returns NULL if the string is NULL.
 *	XnExtendString - Extends 'string' to include 'extension'.  If the extenstion is NULL
 *				the function reduces to XnDupString(string).
 *	XnLowerString - Returns the string all lowercase.  Does not change original.
 *	XnCleanString - Removes the leading and trailing spaces and tabs from 'string'.
 *				Does change the initial contents of the string.
 *	XnSplitString - Splits a string into sections delimited by 'delimeter' and returns an
 *				array of the sections.  The original string is unaffected
 *				and the new sections have their own memory allocated.
 *	XnCreateXmString - Returns an XmString created with the XmSTRING_DEFAULT_CHARSET.  It returns NULL if
 *				the string is NULL.
 *	XnGetXmString - Returns the character string in 'string' using XmStringGetLtoR(..., XmSTRING_DEFAULT_CHARSET).
 *	XnStringArrayToXmString  - Returns an XmString created from the character array 'cs' with 'n' elements.  
 *	XnMallocString - Malloc's a string of n characters.
 *	XnMallocStringTable - Calloc's an array of n Strings.
 */
extern int		XnStrnum			( String string, String pattern );
extern String 		XnLowerString			( String string );
extern String 		XnDupString			( String string );
extern String 		XnExtendString			( String string, String extension );
extern String 		XnCleanString			( String string );
extern String 		XnGetXmString			( XmString string );
extern StringTable	XnSplitString			( String string, String delimeter, int *numsplits,
								Boolean splitquotes );
extern XmString		XnCreateXmString		( String string );
extern XmString		XnStringArrayToXmString		( StringTable cs, int n );
extern String		XnMallocString			( int n );
extern StringTable	XnMallocStringTable		( int n );

/*
 * XnRetrieveUserInfo returns a pointer to a structure of type XnUserInfoStruct which is filled with the
 * following information on the user running the current application.  NULL is returned if an
 * error occurs. For more information on what is in the structure see the appropriate man page.
 *
 * uid		-getuid()
 * pw		-getpwuid(uid)
 * homeDir	-The user's home directory.
 * userName	-pw->pw_name
 * hostname	-gethostname
 * wd		-getwd
 * timeString	-ctime	('\n' is removed)
 * timeRec	-localtime
 */
typedef struct _XnUserInfoStruct XnUserInfoStruct;
struct _XnUserInfoStruct {
	String		homeDir;	/* The user's home directory. */
	String		userName;	/* The user's login name */
	String		hostname;	/* The user's hostname */
	String		timeString;	/* The current time as a string */
	String		wd;		/* The current working directory */
  	uid_t		uid;		/* The user's id number */
	time_t		time;		/* The current time in seconds. */
	struct tm	*timeRec;	/* The current time as a tm struct */
	struct passwd	*pw;		/* The passwd structure */
	};

extern XnUserInfoStruct *	XnRetrieveUserInfo	(void);

/*
 * XnParsePath takes as arguments the following:
 *
 * fullPath	-the path of the file to parse
 * path		-returns a pointer to the path of the file
 * file		-returns a pointer to the filename
 * ext		-returns a pointer to the file extension
 *
 * If any of the fields are missing in the fullPath then its
 * appropriate pointer contains the NULL character (\0).
 * The function returns the complete filename of the file
 * (with extension) or NULL if an error occurs.
 *
 * The path contains the last backslash (/) and the extension does NOT contain the period (.)  
 */
extern String 			XnParsePath	( String fullPath, String *path, String *file, String *ext );

/* 
 * XnRetrieveScreenInfo returns a pointer to a structure
 * of type XnScreenInfoStruct which is filled with 
 * information on the screen associated with 'widget'.
 * It returns NULL on error.  For more information
 * on the structure contents refer to the appropriate man page.
 *
 * display		-XtDisplay(widget)
 * defaultScreen	-DefaultScreen(display)
 * rootWindow		-RootWindow(display, defaultScreen)
 * depth		-DefaultDepth(display, defaultScreen)
 * fg			-XtGetValues on XmNforeground
 * bg			-XtGetValues on XmNbackground
 * blackPixel		-BlackPixel(display, defaultScreen)
 * whitePixel		-WhitePixel(display, defaultScreen)
 */
typedef struct _XnScreenInfoStruct XnScreenInfoStruct;
struct _XnScreenInfoStruct {
	Display		*display;	/* The display */
	int		defaultScreen;	/* The default screen */
	Window		rootWindow;	/* The root window */
	unsigned int	depth;		/* The screen depth */
	unsigned long	fg;		/* The current foreground */
	unsigned long	bg;		/* The current background */
	unsigned long	blackPixel;	/* The black pixel for the screen */
	unsigned long	whitePixel;	/* The white pixel for the screen */
	};

extern XnScreenInfoStruct *	XnRetrieveScreenInfo	( Widget widget );

/*
 * XnIsColorDisplay returns 0 if the current display is monochrome.
 * It returns the depth of the screen if it is color, and it returns
 * a negative value on error.
 */
extern int			XnIsColorDisplay	( void );

/*
 * XnErrorMessage allows a program to output error messages or
 * warnings to 'stderr'.  The format is as follows:
 *
 *	XnErrorMessage( "...", String app_name, int err_type, String proc_name, ... );
 *
 * Where the first argument "..." is a variable argument list that does the same
 * type conversions as 'printf'.  "app_name" is the name of the application.
 * "err_type" is one of XnFATAL, XnWARNING, XnERROR.  "proc_name" is the name
 * of the procedure that triggered the message. The rest of the arguments "..."
 * are the arguments to be converted in the first "...".
 *
 * The output is as follows:
 *	app_name: ERROR : (proc_name) : <message>.
 * 
 * If the error type is XnFATAL the application will exit.
 */
#define	XnERROR		1000		/* ERROR. */
#define	XnWARNING	1001		/* WARNING. */
#define	XnFATAL		1002		/* FATAL - exits immediately. */

#if (!defined __CENTERLINE__) && (!defined hpux)
#include <varargs.h>
extern void			XnErrorMessage	();
#else
extern void			XnErrorMessage	( String comment, String appName, int errorType, String functionName );
#endif

/*
 * Returns the string from 'strerror' unless the system dosen't
 * provide one...  Then it provides a generic message (for now).
 */
extern String XnStrerror	(int errorno);

/*
 * Functions that manipulate the List widget:
 *	XnListAddSortedXmStringItem - Adds an XmString item to 'list' and sorts the list items.
 *	XnListAddSortedXmStringItems - Adds a list of XmString items to 'list' and sorts the list items.
 *	XnListGetStringItems - Returns an array of pointers to character strings of the items in a list.
 *	XnListSetXmStringItems - Sets the items in a list.
 *	XnListGetSingleSelectedStringItem - Returns the single selected item from a list.
 *	XnListItemCount - Returns the number of items in a list.
 *	XnListStringItemExists - Returns True if the specified item exists in the list.
 *	XnListDeleteStringItem - Deletes specified item from list.
 */
extern void		XnListAddSortedXmStringItem		( Widget list, XmString item );
extern void		XnListAddSortedXmStringItems		( Widget list, XmStringTable items, int itemCount );
extern StringTable	XnListGetStringItems			( Widget list, int *numItems );
extern void		XnListSetXmStringItems			( Widget list, XmStringTable items, int itemCount );
extern String		XnListGetSingleSelectedStringItem	( Widget list );
extern int		XnListItemCount				( Widget list );
extern Boolean		XnListStringItemExists			( Widget list, String item );
extern void		XnListDeleteStringItem			( Widget list, String item );

/*
 * Functions that manipulate the user data for widgets:
 *
 *	XnGetUserDataPointer - Returns the POINTER of the XmNuserData of 'widget'.
 *	XnGetUserDataInt - Returns the integer value STORED IN the XmNuserData of 'widget'.
 *	XnGetUserDataFloat - Returns the floating point value STORED IN the XmNuserData of 'widget'.
 *
 *	XnSetUserDataPointer - Sets the POINTER of the XmNuserData of 'widget' to 'data'.
 *	XnSetUserDataInt - Sets the integer value STORED IN the XmNuserData of 'widget' to 'data'.
 *	XnSetUserDataFloat - Sets the floating point value STORED IN the XmNuserData of 'widget' to 'data'.
 */
extern XtPointer	XnGetUserDataPointer	( Widget widget );
extern int		XnGetUserDataInt	( Widget widget );
extern float		XnGetUserDataFloat	( Widget widget );

extern void		XnSetUserDataPointer	( Widget widget, XtPointer data );
extern void		XnSetUserDataInt	( Widget widget, int data );
extern void		XnSetUserDataFloat	( Widget widget, float data );


/*
 * The following functions manipulate application configuration files that are
 * modeled after the Windows '.ini' file scheme.  The configuration files contain
 * sections and each section contains keys.  For example:
 *
 *	# This line is a comment.
 *	# This line is a comment.
 *
 *	[ Section 1 ]
 *	key1 : True	
 *	key2 : 1.5
 *
 *	[ Section 2 ]
 *	key1 : 8
 *	key2 : some string value
 *
 * The functions below allow an application to store and retrieve values associated
 * with a section/key combination.  Key values can be floats, integers, Boolean, and
 * strings.  They are always stored in the file as ASCII character strings. 
 * The configuration file associated with an application is kept in the users HOME
 * directory as a "dotfile", and is based on the application name.  For example
 * the configuration file for "NXpr" would be "$HOME/.nxprrc".  The filename is
 * all lowercase.  This file is constructed with the XnCreateIniFilename function.
 *
 * When setting a value for a key, if the section/key does not exist, they will be created.
 * In all functions, specifying a NULL section implies the FIRST section of keys in the file.
 *
 * Along with the tradition of the Resource Database and Command Line Argument parsing
 * the XnIniLoadResources function will allow you to specify an XtResourceList to fill with
 * values from an ini file.  The method is the same as loading resources from the
 * Resource Database with XtResource structures.  Below is an example declaration:
 *
 * XtResource resources[]=
 * {
 *	{"key1",	"Section 1",	XtRFloat,	sizeof(float),		XtOffset(struct.val1),	XtRFloat,	(XtPointer)NULL},
 *	{"key2",	"Section 2",	XtRString,	sizeof(String),		XtOffset(struct.val2),	XtRString,	(XtPointer)NULL}
 * }
 * 
 * The resource definitions in the structure are (in order):
 *
 *	String      section_key		* Resource key                             *
 *	String      section_name	* Resource section                         *
 *	String      resource_type	* Representation type desired              *
 *	Cardinal    resource_size	* Size in bytes of representation          *
 *	Cardinal    resource_offset	* Offset from base to put resource value   *
 *	String      default_type	* representation type of specified default *
 *	XtPointer   default_addr	* Address of default resource              *
 *
 * Notice the similarity? The KEY parallels the resource name and the SECTION parallels the CLASS.
 *
 * Currently valid resource/default types are:
 *
 *	XtRString
 *	XtRFloat
 *	XtRInt
 *	XtRBoolean
 *
 * The specifications (e.g. if default value is NULL the resource isn't changed)
 * are the same as calling the individual XnIni functions.
 *
 * When specifing strings as a resource type you must supply a POINTER, not a static array.
 *
 */

/*
 * The XnIni functions that return integers will return one of the
 * following values...
 */
#define INI_SUCCESS		0	/* The operation was successful. */
#define INI_NO_KEY		101	/* The specified key does not exist. */
#define INI_NO_SECTION		100	/* The specified section does not exist. */
#define	INI_INVALID_FILE	103	/* A file operation was denied. */
#define INI_INVALID_VALUE	102	/* A parameter or value was invalid. */

/*
 * Creates and returns an ini dotfile based on the name of the application.
 */
extern String 		XnCreateIniFilename		( String appname );

/*
 * Appends a comment to the end of the ini file.
 */
extern int		XnIniWriteComment		( String filename, String comment );

/*
 * These functions return a list of sections/keys that are in the file.
 */
extern StringTable	XnIniGetSectionList		( String filename, int *count );
extern StringTable	XnIniGetSectionKeyList		( String filename, String section, int *count );

/*
 * These functions GET the value for a section/key combination.
 * The Int, String, Float, and Boolean functions retrieve the value associated with a section/key
 * combination as an integer, string, floating point value, and boolean value respectively.
 * The StringList function retrieves the vector (array) of string values associated with a key.
 * For a description of the XnIniLoadResources function see above.
 */
extern int		XnIniGetSectionKeyInt		( String filename, String section, String key, int *value, int *defaultValue );
extern int		XnIniGetSectionKeyString	( String filename, String section, String key, String *value, String defaultValue );
extern int		XnIniGetSectionKeyFloat		( String filename, String section, String key, float *value, float *defaultValue );
extern int		XnIniGetSectionKeyBoolean	( String filename, String section, String key, Boolean *value, Boolean *defaultValue );
extern int		XnIniGetSectionKeyStringList	( String filename, String section, String key, StringTable *value, StringTable *defaultValue );

extern Boolean		XnIniLoadResources		( String filename, XtPointer base, XtResourceList resources, Cardinal num_resources );

/*
 * These functions SET the value for a section/key combination.
 * The Int, String, Float, and Boolean functions save the value associated with a section/key
 * combination as an integer, string, floating point value, and boolean value respectively.
 * The StringList function saves a vector (array) of string values with a key.
 */
extern int		XnIniSetSectionKeyInt		( String filename, String section, String key, int value );
extern int		XnIniSetSectionKeyString	( String filename, String section, String key, String  value );
extern int		XnIniSetSectionKeyFloat		( String filename, String section, String key, float value );
extern int		XnIniSetSectionKeyBoolean	( String filename, String section, String key, Boolean value );
extern int		XnIniSetSectionKeyStringList	( String filename, String section, String key, StringTable value );

/*
 * The XnExecCommand will execute a command and return the output from the
 * command in strings.  The 'command' and 'argv' arguments are for the 'exec'
 * system call.  The stdinfile, stdoutfile, and stderrfile are filenames for
 * redirection of stdin, stdout, and stderr.  If the argument is NULL then
 * /dev/null is used as the redirection file.  If the argument is empty ("") then
 * no redirection takes place.  The stdinput is a string that can be used as
 * redirected input instead of stdinfile.  They are mutually exclusive and one
 * must be NULL.  The stdoutput and stderrput are strings that will contain the
 * stdout and stderr output after the command is finished.  The exitCode is a
 * return value from wait'ing on the command.  It will be 127 if the command could
 * not be executed at all or another value if the command failed.  The command
 * can be run in the background (not waited on) if 'background' is True, and
 * if the removeFiles flag is True then the 3 stdinfile, stdoutfile, and stderrfile
 * will be removed when the command exits.  The output from the command can still
 * be retrieved even if the 3 files aren't supplied.  The function will create
 * temporary filenames for redirection and read the output from those files.
 * If no redirection is used (the filename is "") the output will not be read into
 * a string.
 */
extern Boolean		XnExecCommand( String command, String argv[],
				String stdinfile, String stdoutfile, String stderrfile,
				String stdinput, String *stdoutput, String *stderrput,
				int *exitCode, Boolean background, Boolean removeFiles );

#endif /* _NX_UTILS_H_ */

