/*
 *  NXUtils - nxclipboard.c 
 *
 *	Version 2.0
 *      by Lance Lovette
 *	Code and Display Copyright (C) 1993 by Lance Lovette
 *	All Rights Reserved
 *
 *	Lance Lovette (lovette@iftech.com)
 *	North Carolina State University
 *	Engineering Computer Operations
 *	Department of Computer Science
 *	Raleigh, NC 27607
 *
 *  Permission to use, copy, modify and distribute this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both the
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the names of Lance Lovette and North Carolina
 *  State University not be used in advertising or publicity pertaining to
 *  distribution of the software without specific, written prior permission.
 *  Lance Lovette and North Carolina State University make no
 *  representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 *  Lance Lovette AND North Carolina State University DISCLAIMS ALL
 *  WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Lance Lovette OR
 *  North Carolina State University BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 *  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 *  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 *  OF THIS SOFTWARE.
 *
 *  $RCSfile: nxclipboard.c,v $
 *  $Revision: 2.2 $
 *  $Date: 1993/12/10 03:03:39 $
 */
static char rcsid[] = "$Id: nxclipboard.c,v 2.2 1993/12/10 03:03:39 ltlovett Exp $";

#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/CutPaste.h>

#include "NXUtils.h"

/********************************************************************/

/*
 *  Proc:       XnRetrieveClipboardContents
 *  Callers:    Unknown
 *  Purpose:    Retrieve and return the contents of the clipboard with the specified
 *		format.  If any errors occur, NULL is returned.  If the clipboard is
 *		empty, the empty string ("") is returned.
 *  Design:     01-26-93
 *  Params:     widget     -A widget to use to find the display.
 *		timestamp  -A current timestamp (from an event) for XmClipboard...
 *		format     -The format of the clipboard to retrieve.
 *		return_len -The length of the returned string (Return Value).
 *  Globals:    None
 */
String XnRetrieveClipboardContents(Widget widget, Time timestamp, String format, unsigned long *return_len)
{
	int		private_id	=0;
	String		buffer		=NULL;
	Display		*display	=NULL;
	Window		window		=(Window)NULL;
	unsigned long	num_bytes	=0;
	unsigned long	buf_length	=0;

	(*return_len) = 0;
	
	window = XtWindow( widget );
	display = XtDisplay( widget );
	
	if( XmClipboardInquireLength( display, window, format, &buf_length ) != ClipboardSuccess ) return( NULL );
	if( XmClipboardStartRetrieve( display, window, timestamp ) != ClipboardSuccess ) return( NULL );
	
	buffer = XnMallocString( buf_length );
	
	if( XmClipboardRetrieve( display, window, format, buffer, buf_length, &num_bytes, &private_id ) != ClipboardSuccess )
	{
		XnFreeString( buffer );
		return( NULL );
	}
	
	if( XmClipboardEndRetrieve( display, window ) != ClipboardSuccess )
	{
		XnFreeString( buffer );
		return( NULL );
	}
	
	(*return_len) = num_bytes;
	buffer[num_bytes - 1] = '\0';

	return( buffer );
}

/********************************************************************/

