/*
 *  NXUtils - nxutils.c 
 *
 *	Version 2.0
 *      by Lance Lovette
 *	Code and Display Copyright (C) 1993 by Lance Lovette
 *	All Rights Reserved
 *
 *	Lance Lovette (lovette@iftech.com)
 *	North Carolina State University
 *	Engineering Computer Operations
 *	Department of Computer Science
 *	Raleigh, NC 27607
 *
 *  Permission to use, copy, modify and distribute this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both the
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the names of Lance Lovette and North Carolina
 *  State University not be used in advertising or publicity pertaining to
 *  distribution of the software without specific, written prior permission.
 *  Lance Lovette and North Carolina State University make no
 *  representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 *  Lance Lovette AND North Carolina State University DISCLAIMS ALL
 *  WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Lance Lovette OR
 *  North Carolina State University BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 *  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 *  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 *  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 *  OF THIS SOFTWARE.
 *
 *  $RCSfile: nxutils.c,v $
 *  $Revision: 2.5 $
 *  $Date: 1994/02/27 18:21:41 $
 */
static char rcsid[] = "$Id: nxutils.c,v 2.5 1994/02/27 18:21:41 reward Exp $";

#include <ctype.h>
#include <sys/types.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <strings.h>

#include <Xm/List.h>
#include <Xm/Text.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>
#include <Xm/CascadeB.h>
#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>

#include <X11/Intrinsic.h>

#include "NXUtils.h"

/*******************************************************/

/*
 *  Proc:	XnFileSize
 *  Purpose:	Determine size of a file.
 *  Params:	filename - The name of the file.
 *  Returns:	The size of the file or -1 on error.
 */
int XnFileSize( String filename )
{
	struct stat stbuf;

	if( !filename ) return( -1 );

	if( stat(filename, &stbuf) >= 0 )
		return( stbuf.st_size );

	return( -1 );
}

/*******************************************************/

/*
 *  Proc:	XnDeleteFile
 *  Purpose:	Deletes a file or directory.
 *  Params:	filename - The name of the file/directory.
 *  Returns:	True if the file/directory is deleted.
 */
Boolean XnDeleteFile( String filename )
{
	if( !filename ) return( False );

	if( remove(filename) == 0 )
		return( True );

	return( False );
}

/*******************************************************/

/*
 *  Proc:	XnFileReadable
 *  Purpose:	Determines if a file is readable.
 *  Params:	filename - The name of the file.
 *  Returns:	True if the file is readable.
 */
Boolean XnFileReadable( String filename )
{
	if( !filename ) return( False );

	if( access(filename, R_OK) == 0 )
		return( True );

	return( False );
}

/*******************************************************/

/*
 *  Proc:	XnFileWritable
 *  Purpose:	Determines if a file is writable.
 *  Params:	filename - The name of the file.
 *  Returns:	True if the file is writable.
 */
Boolean XnFileWritable( String filename )
{
	FILE	*file	=NULL;

	if( !filename ) return( False );

	if( access(filename, W_OK) == 0 )
		return( True );

	if( !XnFileExists(filename) )
	{
		file = fopen( filename, "w" );
		if( file )
		{
			fclose( file );
			XnDeleteFile( filename );

			return( True );
		}
	}

	return( False );
}

/*******************************************************/

/*
 *  Proc:	XnFileExists
 *  Purpose:	Determines if a file exists.
 *  Params:	filename - The name of the file.
 *  Returns:	True if the file exists.
 */
Boolean XnFileExists( String filename )
{
	if( !filename ) return( False );

	if( access(filename, F_OK) == 0 )
		return( True );

	return( False );
}

/*******************************************************/

/*
 *  Proc:	XnMakeLabel
 *  Purpose:	Create a managed label.
 *  Params:	parent - The parent of the label.
 *		name - The name for the label.
 *		labelString - The label string.
 *		argList - Extra arguments. 
 *		argCount - The count of arguments.
 *  Returns:	The managed label widget.
 */
Widget XnMakeLabel( Widget parent, String name, String labelString, Arg *argList, int argCount )
{
	Arg		al[1];
	Widget		label;
	XmString	message;

	message = XnCreateXmString( labelString );
	XtSetArg( al[0],	XmNlabelString,		message );
	label = XmCreateLabel( parent, name, al, 1 );
	XtManageChild( label );
	XmStringFree( message );

	if( argList && argCount ) XtSetValues( label, argList, argCount );

	return( label );
}

/*******************************************************/

/*
 *  Proc:	XnAddAccelerator
 *  Purpose:	Adds an accelerator to a widget.
 *  Params:	widget - The widget.
 *		acceleratorText - The accelerator text.
 *		acceleratorKey - The accelerator key.
 *  Returns:	True if successful.
 */
Boolean XnAddAccelerator( Widget widget, String acceleratorText, String acceleratorKey )
{
	Arg	al[2];

	XtSetArg( al[0],	XmNacceleratorText,	XmStringCreate(acceleratorText, XmSTRING_DEFAULT_CHARSET) );
	XtSetArg( al[1],	XmNaccelerator,		acceleratorKey );
	XtSetValues( widget, al, 2 );

	return( True );
}

/*******************************************************/

/*
 *  Proc:	XnMakeMenu
 *  Purpose:	Creates a pulldown menu with a button.  The button
 *		is a cascade button.  The name of the button is the name
 *		of the menu (menuName) with 'Button' on the end.
 *  Params:	parent - The parent of the menu (usually a menu bar).
 *		menuName - The name of the menu.
 *		menuLabel - The label for the menu.
 *		mnemonic - The mnemonic for the menu.
 *  Returns:	The menu widget.
 */
Widget XnMakeMenu( Widget parent, String menuName, String menuLabel, KeySym mnemonic )
{
	Arg		al[3];
	String		buttonName;
	Widget		menu;
	Widget		cascade;
	XmString	message;

	menu = XmCreatePulldownMenu( parent, menuName, NULL, 0 );

	message = XnCreateXmString( menuLabel );
	buttonName = XnMallocString( strlen(menuName) + 8 );
	sprintf( buttonName, "%sButton", menuName );

	XtSetArg( al[0],	XmNsubMenuId,	menu );
	XtSetArg( al[1],	XmNmnemonic,	mnemonic );
	XtSetArg( al[2],	XmNlabelString,	message );
	cascade = XmCreateCascadeButton( parent, buttonName, al, 3 );
	XtManageChild( cascade );

	XnFreeString( buttonName );
	XmStringFree( message );

	return( menu );
}

/*******************************************************/

/*
 *  Proc:	XnMakeHelpMenu
 *  Purpose:	Creates a pull-down menu that is designated as the
 *		'help' menu in a menu-bar.  See XnMakeMenu.
 *  Params:	See XnMakeMenu.
 *  Returns:	The menu widget.
 */
Widget XnMakeHelpMenu( Widget parent, String menuName, String menuLabel, KeySym mnemonic )
{
	Arg		al[3];
	String		buttonName;
	Widget		menu;
	Widget		cascade;
	XmString	message;

	menu = XmCreatePulldownMenu( parent, menuName, NULL, 0 );

	message = XnCreateXmString( menuLabel );
	buttonName = XnMallocString( strlen(menuName) + 8 );
	sprintf( buttonName, "%sButton", menuName );

	XtSetArg( al[0],	XmNsubMenuId,		menu );
	XtSetArg( al[1],	XmNmnemonic,		mnemonic );
	XtSetArg( al[2],	XmNlabelString,		message );
	cascade = XmCreateCascadeButton( parent, buttonName, al, 3 );
	XtManageChild( cascade );

	XnFreeString( buttonName );
	XmStringFree( message );


	XtSetArg( al[0],	XmNmenuHelpWidget,	cascade );
	XtSetValues( parent, al, 1 );

	return( menu );
}

/*******************************************************/

/*
 *  Proc:	XnMakeMenuOption
 *  Purpose:	Creates a push button that is an option in a pulldown menu.
 *  Params:	parent - The parent of the button (the menu).
 *		optionName - The name for the button. 
 *		optionLabel - The label string for the button.
 *		mnemonic - The mnemonic of the button.
 *		callback - The callback for the button.
 *		clientData - Client data for the button callback.
 *  Returns:	The managed button.
 */
Widget XnMakeMenuOption( Widget parent, String optionName, String optionLabel, KeySym mnemonic, 
				void (*callback)(Widget, XtPointer, XmPushButtonCallbackStruct *),
				XtPointer clientData )
{
	Arg		al[2];
	Widget		button;
	XmString	message;

	message = XnCreateXmString( optionLabel );

	XtSetArg( al[0],	XmNlabelString,		message );
	XtSetArg( al[1],	XmNmnemonic,		mnemonic );
	button = XmCreatePushButton( parent, optionName, al, 2 );

	if( callback ) XtAddCallback( button, XmNactivateCallback, (XtCallbackProc)callback, clientData );

	XtManageChild( button );

	XmStringFree( message );

	return( button );
}

/*******************************************************/

/*
 *  Proc:	XnMakeToggleButton
 *  Purpose:	Creates a managed toggle button.
 *  Params:	parent - The parent of the button.
 *		optionName - The name for the button. 
 *		optionLabel - The label string for the button.
 *		mnemonic - The mnemonic of the button.
 *		callback - The callback for the button.
 *		clientData - Client data for the button callback.
 *		argList - Any extra arguments.
 *		argCount - The count of extra arguments.
 *  Returns:	The managed toggle button.
 */
Widget XnMakeToggleButton( Widget parent, String optionName, String optionLabel, 
				void (*callback)(Widget, XtPointer, XmToggleButtonCallbackStruct *),
				XtPointer clientData, Arg *argList, int argCount )
{
	Arg		al[1];
	Widget	button;
	XmString	message;

	message = XnCreateXmString( optionLabel );

	XtSetArg( al[0],	XmNlabelString,		message );
	button = XmCreateToggleButton( parent, optionName, al, 1 );

	if( argList && argCount ) XtSetValues( button, argList, argCount );

	if( callback ) XtAddCallback( button, XmNvalueChangedCallback, (XtCallbackProc)callback, clientData );

	XtManageChild( button );

	XmStringFree( message );

	return( button );
}

/*******************************************************/

/*
 *  Proc:	XnMakePushButton
 *  Purpose:	Creates a managed push button.
 *  Params:	parent - The parent of the button.
 *		optionName - The name for the button. 
 *		optionLabel - The label string for the button.
 *		mnemonic - The mnemonic of the button.
 *		callback - The callback for the button.
 *		clientData - Client data for the button callback.
 *		argList - Any extra arguments.
 *		argCount - The count of extra arguments.
 *  Returns:	The managed push button.
 */
Widget XnMakePushButton( Widget parent, String optionName, String optionLabel, 
				void (*callback)(Widget, XtPointer, XmPushButtonCallbackStruct *),
				XtPointer clientData, Arg *argList, int argCount )
{
	Arg		al[1];
	Widget	button;
	XmString	message;

	message = XnCreateXmString( optionLabel );

	XtSetArg( al[0],	XmNlabelString,		message );
	button = XmCreatePushButton( parent, optionName, al, 1 );

	if( argList && argCount ) XtSetValues( button, argList, argCount );

	if( callback ) XtAddCallback( button, XmNactivateCallback, (XtCallbackProc)callback, clientData );

	XtManageChild( button );

	XmStringFree( message );

	return( button );
}

/*******************************************************/

/*
 *  Proc:	XnFileSelectionBoxCWD
 *  Purpose:	Determine the working directory of a FSB.
 *  Params:	fileSBDialog - The dialog.
 *  Returns:	The directory.
 */
String XnFileSelectionBoxCWD( Widget fileSBDialog )
{
	Arg		al[1];
	String		s		=NULL;
	XmString	directory       =NULL;

	XtSetArg( al[0],       XmNdirectory,	&directory );
	XtGetValues( fileSBDialog, al, 1 );

	XmStringGetLtoR( directory, XmSTRING_DEFAULT_CHARSET, &s );
	XmStringFree( directory );

	return( s );
}

/*******************************************************/

/*
 *  Proc:	XnSetLabelString
 *  Purpose:	Sets the XmNlabelString string of a widget.
 *  Params:	widget - The widget to set label of.
 *		string - The new label string.
 *  Returns:	None
 */
void XnSetLabelString( Widget widget, String string )
{
	Arg		al[1];
	XmString	message;

	message = XnCreateXmString( string );
	XtSetArg( al[0],      XmNlabelString,      message );
	XtSetValues( widget, al, 1 );
	XmStringFree( message );
}

/*******************************************************/

/*
 *  Proc:	XnSetDialogText
 *  Purpose:	Sets the string of the XmDIALOG_TEXT widget of
 *		a selection box dialog if the string is NON-NULL.
 *  Params:	widget - The dialog.
 *		string - The new text for the dialog.
 *  Returns:	The text area of the dialog.
 */
Widget XnSetDialogText( Widget widget, String string )
{
	Widget	edit	=NULL;

	if( widget )
	{
		edit = XmSelectionBoxGetChild( widget, XmDIALOG_TEXT );
		if( edit && string ) XmTextSetString( edit, string );
	}

	return( edit );
}

/*******************************************************/

/*
 *  Proc:	XnGetDialogText
 *  Purpose:	Gets the string of the XmDIALOG_TEXT widget of
 *		a selection box dialog if the string is NON-NULL.
 *  Params:	widget - The dialog.
 *  Returns:	The text in the dialog or NULL if none.
 */
String XnGetDialogText( Widget widget )
{
	String	s	=NULL;
	Widget	edit	=NULL;

	if( widget )
	{
		edit = XmSelectionBoxGetChild( widget, XmDIALOG_TEXT );
		if( edit ) s = XmTextGetString( edit );

		if( s && !*s )
		{
			XnFreeString( s );
			s = NULL;
		}
	}

	return( s );
}

/*******************************************************/

/*
 *  Proc:	XnSetDialogMessageString
 *  Purpose:	Sets the XmNmessageString of a dialog.
 *  Params:	widget - The dialog widget.
 *		string - The new message string.
 *  Returns:	None
 */
void XnSetDialogMessageString( Widget widget, String string )
{
	Arg		al[1];
	XmString	message;

	message = XnCreateXmString( string );
	XtSetArg( al[0],      XmNmessageString,      message );
	XtSetValues( widget, al, 1 );
	XmStringFree( message );
}

/*******************************************************/

/*
 *  Proc:	XnSetDialogType
 *  Purpose:	Sets the XmNdialogType of a dialog.
 *  Params:	widget - The dialog.
 *		type - The new type.
 *  Returns:	None
 */
void XnSetDialogType( Widget widget, unsigned char type )
{
	Arg	al[1];

	XtSetArg( al[0],	XmNdialogType,		type );
	XtSetValues( widget, al, 1 );
}

/*******************************************************/

/*
 *  Proc:	XnCreateXmString
 *  Purpose:	Create an XmString representing a string.
 *  Params:	string - The string to represent.
 *  Returns:	The new XmString representation.
 */
XmString XnCreateXmString( String string )
{
	return( XmStringCreateLtoR(string, XmSTRING_DEFAULT_CHARSET) );
}

/*******************************************************/

/*
 *  Proc:	XnGetXmString
 *  Purpose:	Convert an XmString to a character string.
 *  Params:	string - The XmString to convert.
 *  Returns:	The character string representation.
 */
String XnGetXmString( XmString string )
{
	String	s	=NULL;

	XmStringGetLtoR( string, XmSTRING_DEFAULT_CHARSET, &s );

	return( s );
}

/*******************************************************/

/*
 *  Proc:	XnStringArrayToXmString
 *  Purpose:	Convert an array of character strings into a
 *		single XmString.
 *  Params:	cs - The array of character strings. 
 *		n - The number of strings in the array.
 *  Returns:	The XmString representation.
 */
XmString XnStringArrayToXmString( StringTable cs, int n )
{
	int        i;
	XmString   xmstr;

	if( n <= 0 ) return( XmStringCreate("", XmSTRING_DEFAULT_CHARSET) );

	xmstr = (XmString) NULL;

	for( i = 0; i < n; i++ )
	{
		if (i > 0) xmstr = XmStringConcat( xmstr, XmStringSeparatorCreate() ); 
		xmstr = XmStringConcat( xmstr, XmStringCreate(cs[i], XmSTRING_DEFAULT_CHARSET) );
	}

	return( xmstr );
}

/*******************************************************/

/*
 *  Proc:	XnStrnum
 *  Purpose:	Returns the number of times a specified pattern
 *		appears in a string.
 *  Params:	string - The string. 
 *		pattern - The pattern.
 *  Returns:	The number of times the pattern appears in
 *		the string.
 */
int XnStrnum( String string, String pattern )
{
	int	sl;
	int	pl;
	int	match;

	if( !string || !pattern ) return( 0 );

	sl = strlen( string );
	pl = strlen( pattern );
	match = 0;

	while( sl >= pl )
	{
		if( strncmp(string, pattern, (size_t)pl) == 0 ) match++;
		string++;
		sl = strlen( string );
	}

	return( match );
}

/*******************************************************/

/*
 *  Proc:	XnMallocStringTable
 *  Purpose:	Allocate space for an array of String elements.
 *  Params:	len - The number of array positions to allocate.
 *  Returns:	A pointer to the allocated space.
 */
StringTable XnMallocStringTable( int len )
{
	if( len > 0 )
		return( (StringTable)XtCalloc( len, sizeof(String) ) );

	return( NULL );
}

/*******************************************************/

/*
 *  Proc:	XnMallocString
 *  Purpose:	Allocate space for a String of characters.
 *  Params:	len - The number of characters in the string.
 *  Returns:	A pointer to the allocated space.
 */
String XnMallocString( int len )
{
	if( len > 0 )
		return( (String)XtMalloc( len * sizeof(char) ) );

	return( NULL );
}

/*******************************************************/

/*
 *  Proc:	XnDupString
 *  Purpose:	Duplicate the contents of a string.
 *  Params:	string - The string to duplicate.
 *  Returns:	A copy of the string or NULL if the string
 *		dosen't exist.
 */
String XnDupString( String string )
{
	String	new	=NULL;

	if( !string ) return( NULL );

	new = XnMallocString( strlen(string) + 1 );
	strcpy( new, string );

	return( new );
}

/*******************************************************/

/*
 *  Proc:	XnExtendString
 *  Purpose:	Extend the contents of a string. If the
 *		extenstion is NULL the string is duplicated
 *		but not extended.
 *  Params:	string - The string to extend.
 *		extend - The string to add.
 *  Returns:	The new string.
 */
String XnExtendString( String string, String extend )
{
	String	new	=NULL;

	if( !string ) return( NULL );
	if( !extend ) return( XnDupString(string) );

	new = (String) XtRealloc( string, strlen(string) + strlen(extend) + 1 );

	if( new )
	{
		strcpy( new, string );
		strcat( new, extend );
	}
	else
	{
		strcpy( new, "" );
		XnErrorMessage( "XtRealloc() failed", "NXUtils", XnWARNING, "XnExtendString" );
	}

	return( new );
}

/*******************************************************/

/*
 *  Proc:	XnLowerString
 *  Purpose:	Returns a lowercase version of a string.
 *  Params:	string - The string to convert.
 *  Returns:	A new string that is all lowercase.
 */
String XnLowerString( String string )
{
	int	i;
	String	new	=NULL;

	if( !string ) return( NULL );

	new = XnDupString( string );

	for( i=0;  i < strlen(new); i++ )
		new[ i ] = tolower( new[i] );

	return( new );
}

/*******************************************************/

/*
 *  Proc:	XnCleanString
 *  Purpose:	Removes the leading and trailing spaces
 *		and tabs off a string.  Changes the initial
 *		contents of the string.
 *  Params:	temp - The string to clean.
 *  Returns:	The string.
 */
String XnCleanString( String temp )
{
	String	s	=NULL;

	if( !temp ) return( NULL );

	/*
	 * Remove the leading spaces from 'temp'.
	 */
	s = temp;
	while( (*s) && (((*s) == ' ') || ((*s) == '\t')) ) s++;
	strcpy( temp, s );

	/*
	 * Remove the trailing spaces from 'temp'.
	 */
	s = temp;
	while( (*s) != '\0' ) s++;

	if( s != temp )
	{
		s--;
		while( (s != temp) && (((*s) == ' ') || ((*s) == '\t')) ) s--;
		s++;
	}

	(*s) = '\0';

	return( temp );
}

/*******************************************************/

/*
 *  Proc:	XnSplitString
 *  Purpose:	Split a string into sections separated by a delimeter
 *		and return an array of strings containing the sections.
 *  Params:	string - The string to split.
 *		delimeter - The delimeter string.
 *		numsplits - <Return> The number of sections returned.
 *		splitquotes - True if text is quotes is to be split.
 *  Returns:	An array of sections or NULL on error.
 */
StringTable XnSplitString( String string, String delimeter, int *numsplits, Boolean splitquotes )
{
	String		end		=NULL;
	String		start		=NULL;
	String		newstring	=NULL;
	Boolean		inquote		=False;
	StringTable	split		=NULL;

	if( !string || !delimeter || !numsplits ) return( NULL );

	(*numsplits) = 0;
	inquote = False;

	/*
	 * Make our own copy of the string to edit.
	 */
	newstring = XnDupString( string );

	split = XnMallocStringTable( 2 );

	/*
	 * Run through the string.  Everytime we find the delimeter
	 * copy everything from the beginning of the last section up to the
	 * end of that section.  We watch for quotes so that we don't
	 * break strings in quotes up (if the user specified so). We
	 * assume the quotes are matched.
	 */
	end = newstring;
	start = newstring;

	while( *end )
	{
		if( !splitquotes && (*end == '"') ) inquote = (inquote == True) ? False : True;

		if( !inquote && (strncmp(end,delimeter,strlen(delimeter)) == 0) )
		{
			(*end) = '\0';
			end += strlen( delimeter );
			split = (StringTable)XtRealloc( (String)split, (Cardinal)(sizeof(String) * ((*numsplits) + 1)) );
			split[ (*numsplits)++ ] = XnDupString( start );
			start = end;
		}
		else end++;
	}

	if( start != end )
	{
		split = (StringTable)XtRealloc( (String)split, (Cardinal)(sizeof(String) * ((*numsplits) + 1)) );
		split[ (*numsplits)++ ] = XnDupString( start );
	}

	/*
	 * Add a NULL pointer on the end so this vector can be
	 * passed to a function like exec.
	 */
	split = (StringTable)XtRealloc( (String)split, (Cardinal)(sizeof(String) * ((*numsplits) + 1)) );
	split[ (*numsplits) ] = NULL;

	XnFreeString( newstring );

	return( split );
}

/*******************************************************/

/*
 *  Proc:	XnRetrieveUserInfo
 *  Purpose:	Gathers information about the current user and
 *		returns a structure containing the values.
 *  Params:	None
 *  Returns:	The structure of values or NULL on error.
 */
XnUserInfoStruct *XnRetrieveUserInfo( void )
{
	char			wd[ 256 ];
	char			hostname[ 40 ];
	XnUserInfoStruct	*userInfo 	=NULL;

	/*
	 * Use various functions to get information on the
	 * user and copy them into the structure 'userInfo'.
	 * Return the structure or NULL on error.
	 */
	userInfo = (XnUserInfoStruct *) XtMalloc( sizeof(XnUserInfoStruct) );
	if( !userInfo ) return( NULL );

	userInfo->uid = (uid_t)getuid();
	userInfo->pw = (struct passwd *)getpwuid( userInfo->uid );

	if( userInfo->pw )
		userInfo->userName = XnDupString( userInfo->pw->pw_name );

	if( gethostname(hostname, 40) == 0 )
		userInfo->hostname = XnDupString( hostname );
	else userInfo->hostname = (String)NULL;

	userInfo->wd = XnDupString( getcwd( wd, 256 ) );

	userInfo->time = time( NULL );

	userInfo->timeString = ctime( &userInfo->time );
	userInfo->timeString[ XnEndChar(userInfo->timeString) ] = '\0';

	userInfo->timeRec = localtime( &userInfo->time );

	if( userInfo->pw->pw_dir )
		userInfo->homeDir = XnDupString( userInfo->pw->pw_dir );
	else
	{
		if( getenv("HOME") )
			userInfo->homeDir = XnDupString( getenv("HOME") );
		else
			userInfo->homeDir = XnDupString( "" );
	}

	if( strlen(userInfo->homeDir) )
		if( userInfo->homeDir[ XnEndChar(userInfo->homeDir) ] == '/' )
			userInfo->homeDir[ XnEndChar(userInfo->homeDir) ] = '\0';

	return( userInfo );
}


/*******************************************************/

/*
 *  Proc:	XnParsePath
 *  Purpose:	Parse a complete path into its individual sections
 *		of path, file, and file extension.
 *  Params:	fullPath - The complete path of the file.
 *		path - <Return> The path (with trailing /).
 *		file - <Return> The filename.
 *		ext - <Return> The file extension.
 *  Returns:	The complete filename (with extension) or NULL on error.
 */
String XnParsePath( String fullPath, String *path, String *file, String *ext )
{
	String	s		=NULL;
	String	tmp		=NULL;
	String	complete	=NULL;

	/*
	 * Split the string 'fullPath' into separate parts
	 * 1) The path.
	 * 2) The filename.
	 * 3) The filename extension.
	 * 
	 * Return the filename and return NULL strings for
	 * parts of the path that are invalid (do not exist).
	 */
	if( !fullPath ) return( NULL );

	/*
	 * Start at the end of the fullPath and go backwards
	 * down the string until we hit a '/' or run out of
	 * characters.  Everything before the '/' is the path.
	 * If there is no path, point it to NULL.
	 */
	tmp = XnDupString( fullPath );
	s = tmp + ( strlen(tmp) - 1 );
	while( (s) && (*s) && ((*s) != '/') && (s != tmp) ) s--;

	if( s == tmp )
		(*path) = XnDupString( "" );
	else
	{
		if( (s) && (*s) && ((*s) == '/') )
		{ 
			s++; 
			(*s)='\0'; 
		}
		else (*s)='\0';

		(*path) = XnDupString( tmp );
	}

	XnFreeString( tmp );

	/*
	 * Copy everything after the last '/' into filename so
	 * we can parse it for the filename and extension.
	 * First go back to the last '/'.  Everything after this
	 * is considered the complete filename.
	 */
	tmp = XnDupString( fullPath );
	s = tmp + ( strlen(tmp) - 1 );
	while( (s) && (*s) && ((*s) != '/') && (s != tmp) ) s--;
	if( (s) && (*s) && ((*s) == '/') ) s++;

	complete = XnDupString( s );

	XnFreeString( tmp );

	/*
	 * Go back until we hit a '.' indicating the beginning of the extension.  If there is
	 * no extension, point is to NULL.
	 */
	if( complete && (*complete) )
	{
		tmp = XnDupString(complete);
		s = tmp + ( strlen(tmp) - 1 );
		while( (*s) && ((*s) != '.' ) && ( s != tmp ) ) s--;
		if( s == tmp ) s = NULL;
		else
		{
			if( (s) && (*s) == '.' )
			{ 
				(*s) = '\0'; 
				s++; 
			}
		}
	}

	/*
	 * If 's' points to anything, then there is a filename and an extension.
	 * If it points to NULL then there is no extension or the filename
	 * begins with a '.'
	 */
	if( s )
	{
		(*ext) = XnDupString( s );
		(*file) = XnDupString( tmp );
		XnFreeString( tmp );
	}
	else
	{
		(*ext) = XnDupString( "" );

		if( tmp ) (*file) = XnDupString( tmp );
		else
		{
			(*file) = XnDupString( "" );
		}
	}

	return( complete );
}

/*******************************************************/

/*
 *  Proc:	XnRetrieveScreenInfo
 *  Purpose:	Gathers information about the current screen and
 *		returns the values in a structure.
 *  Params:	widget - A widget to determine the values from.
 *  Returns:	The structure of values or NULL on error.
 */
XnScreenInfoStruct *XnRetrieveScreenInfo( Widget widget )
{
	int			defaultScreen;
	Arg			al[2];
	Window			rootWindow;
	Display			*display;
	unsigned int		depth;
	unsigned long		fg;
	unsigned long		bg;
	unsigned long		blackPixel;
	unsigned long		whitePixel;
	XnScreenInfoStruct	*screenInfo;

	/*
	 * Use various functions to get information
	 * on the screen associated with 'widget' and
	 * fill the structure 'screenInfo' with them.
	 * Return NULL on error.
	 */
	if( !widget ) return( NULL );

	display = XtDisplay( widget );
	defaultScreen = DefaultScreen( display );
	rootWindow = RootWindow( display, defaultScreen );
	depth = DefaultDepth( display, defaultScreen );

	XtSetArg( al[0],	XmNforeground,		&fg );
	XtSetArg( al[1],	XmNbackground,		&bg );
	XtGetValues( widget, al, 2 );

	blackPixel = BlackPixel( display, defaultScreen );
	whitePixel = WhitePixel( display, defaultScreen );

	screenInfo = XtNew( XnScreenInfoStruct );

	screenInfo->fg = (unsigned long)fg;
	screenInfo->bg = (unsigned long)bg;
	screenInfo->depth = (unsigned int)depth;
	screenInfo->display = display;
	screenInfo->blackPixel = (unsigned long)blackPixel;
	screenInfo->whitePixel = (unsigned long)whitePixel;
	screenInfo->rootWindow = rootWindow;
	screenInfo->defaultScreen = defaultScreen;

	return( screenInfo );
}

/*******************************************************/

/*
 *  Proc:	XnIsColorDisplay
 *  Purpose:	Determine if the current display is color/mono.
 *  Params:	None
 *  Returns:	0 if mono, depth of screen if color, -1 on error.
 */
int XnIsColorDisplay( void )
{
	int	depth		=0;
	Display	*display	=NULL;

	display = XOpenDisplay( "" );

	if( !display ) return( -1 );

	depth = DefaultDepthOfScreen( DefaultScreenOfDisplay(display) );

	XCloseDisplay( display );

	if( depth > 1 ) return( depth );

	return( 0 );
}

/*******************************************************/

/*
 *  Proc:	XnErrorMessage
 *  Purpose:	An error message function - see NXUtils.h file.
 *  Params:	Variable.
 *  Returns:	None
 */
#if (!defined __CENTERLINE__) && (!defined hpux)
void XnErrorMessage( va_alist )
va_dcl
{
	int		ival		=0;
	int		errorType	=0;
	String		t		=NULL;
	String		s		=NULL;
	String		p		=NULL;
	String		sval		=NULL;
	String		appName		=NULL;
	String		comment		=NULL;
	String		errorName	=NULL;
	String		functionName	=NULL;
	String		formatStr	=NULL;
	double	dval		=0.0;
	va_list	vaList;

	va_start( vaList );

	formatStr = va_arg( vaList, String );
	appName = va_arg(vaList, String);

	errorType = va_arg(vaList, int);
	if( errorType == XnERROR   ) errorName = XnDupString( "ERROR" ); else
	if( errorType == XnWARNING ) errorName = XnDupString( "WARNING" ); else
	if( errorType == XnFATAL   ) errorName = XnDupString( "FATAL"); else
	errorName = XnDupString("NONE");

	functionName = va_arg( vaList, String );

	s = XnMallocString( 200 );
	comment = XnMallocString( 500 );

	for( t=comment, p=formatStr; *p; p++ )
	{
		if( *p != '%' )
		{
			(*t) = (*p);
			t++;
			continue;
		}

		switch( *++p )
		{
		case 'd':
			ival = va_arg(vaList, int);
			sprintf( s, "%d", ival );
			for( ; *s != '\0'; s++, t++ ) (*t) = (*s);
			break;

		case 'f':
			dval = va_arg( vaList, double );
			sprintf( s, "%f", dval );
			for( ; *s != '\0'; s++, t++ ) (*t) = (*s);
			break;

		case 's':
			for( sval=va_arg(vaList, String); *sval; sval++, t++ )
				(*t) = (*sval);
			break;

		default:
			(*t) = (*p);
			break;
		}
	}
	(*t) = '\0';

	va_end( vaList );

	if( *functionName )
		fprintf( stderr, "%s: %s (%s) %s.\n", appName, errorName, functionName, comment );
	else
		fprintf( stderr, "%s: %s : %s.\n", appName, errorName, comment );

	XnFreeString( comment );

	if( errorType == XnFATAL ) exit( 1 );
}
#else
void XnErrorMessage(String comment, String appName, int errorType, String functionName)
{
	String	errorName	=NULL;

	if( errorType == XnERROR   ) errorName = XnDupString( "ERROR" ); else
	if( errorType == XnWARNING ) errorName = XnDupString( "WARNING" ); else
	if( errorType == XnFATAL   ) errorName = XnDupString( "FATAL"); else
	errorName = XnDupString( "NONE" );

	fprintf( stderr, "%s: %s (%s) %s.\n", appName, errorName, functionName, comment );

	if( errorType == XnFATAL ) exit( 1 );
}
#endif

/*******************************************************/

/*
 *  Proc:	XnListStringItemExists
 *  Purpose:	Determines if a string exists in a list widget.
 *  Params:	list - The list widget.
 *		string - The item to search for.
 *  Returns:	True if the item exists.
 */
Boolean XnListStringItemExists( Widget list, String string )
{
	XmString	item;
	Boolean		exists	=False;

	if( list && string )
	{
		item = XnCreateXmString( string );

		if( item )
		{
			exists = XmListItemExists( list, item );
			XmStringFree( item );
		}
	}

	return( exists );
}

/*******************************************************/

/*
 *  Proc:	XnListDeleteStringItem
 *  Purpose:	Delete the specified item from a list.
 *  Params:	list - The list widget.
 *		string - The item to delete.
 *  Returns:	None
 */
void XnListDeleteStringItem( Widget list, String string )
{
	XmString	item;

	if( list && string )
	{
		item = XnCreateXmString( string );

		if( item )
		{
			XmListDeleteItem( list, item );
			XmStringFree( item );
		}
	}
}

/*******************************************************/

/*
 *  Proc:	XnListGetSingleSelectedStringItem
 *  Purpose:	Determines the single selected item in a list widget.
 *  Params:	list - The list widget.
 *  Returns:	The selected item or NULL if nothing is selected.
 */
String XnListGetSingleSelectedStringItem( Widget list )
{
	int		numItems	=0;
	Arg		al[2];
	XmStringTable	items		=NULL;

	if( list )
	{
		XtSetArg( al[0],		XmNselectedItems,		&items );
		XtSetArg( al[1],		XmNselectedItemCount,		&numItems );
		XtGetValues( list, al, 2 );
	}

	if( !items || (numItems != 1) ) return( NULL );

	return( XnGetXmString(items[0]) );
}

/*******************************************************/

/*
 *  Proc:	XnListAddSortedXmStringItemXmStringCMP
 *  Purpose:	Compare two XmStrings.
 *  Params:	a, b - The strings to compare.
 *  Returns:	0 if equal - like strcmp.
 */
#if (defined _INCLUDE_HPUX_SOURCE) && (!defined __CENTERLINE__)
int XnListAddSortedXmStringItemXmStringCMP( const void *a, const void *b )
#else
int XnListAddSortedXmStringItemXmStringCMP( void *a, void *b )
#endif
{
	String	s1	=NULL;
	String	s2	=NULL;

	s1 = XnGetXmString( *(XmString *)a );
	s2 = XnGetXmString( *(XmString *)b );

	return( strcmp( s1, s2 ) );
}

/*******************************************************/

/*
 *  Proc:	XnListAddSortedXmStringItem
 *  Purpose:	Adds an item into a list and sorts the list.
 *  Params:	list - The list widget.
 *		addItem - The new item to add.
 *  Returns:	None.
 */
void XnListAddSortedXmStringItem( Widget list, XmString addItem )
{
	int		i;
	int		numItems	=0;
	Arg		al[2];
	XmStringTable	items		=NULL;
	XmStringTable	newItems	=NULL;

	if( !list || !addItem ) return;

	XtSetArg( al[0],		XmNitems,		&items );
	XtSetArg( al[1],		XmNitemCount,		&numItems );
	XtGetValues( list, al, 2 );

	newItems = (XmStringTable) XtMalloc( sizeof(XmString) * (numItems + 1) );
	for( i=0; i < numItems; i++ )
		newItems[i] = XmStringCopy( items[i] );
	newItems[i] = XmStringCopy( addItem );

	qsort( (void *)newItems, (size_t)numItems + 1, sizeof(XmString), XnListAddSortedXmStringItemXmStringCMP );

	XtSetArg( al[0],		XmNitems,		newItems );
	XtSetArg( al[1],		XmNitemCount,		numItems + 1);
	XtSetValues( list, al, 2 );
}

/*******************************************************/

/*
 *  Proc:	XnListAddSortedXmStringItems
 *  Purpose:	Adds multiple items to a list and sorts the list.
 *  Params:	list - The list widget.
 *		addItems - The items to add.
 *		numAdds - The number of items to add.
 *  Returns:	None
 */
void XnListAddSortedXmStringItems( Widget list, XmStringTable addItems, int numAdds )
{
	int		i;
	int		j;
	int		numOld		=0;
	Arg		al[2];
	XmStringTable	oldItems	=NULL;
	XmStringTable	newItems	=NULL;

	if( !list || !addItems || !numAdds ) return;

	XtSetArg( al[0],		XmNitems,		&oldItems );
	XtSetArg( al[1],		XmNitemCount,		&numOld );
	XtGetValues( list, al, 2 );

	newItems = (XmStringTable) XtCalloc( numOld + numAdds, sizeof(XmString) );

	for( i=0; i < numOld; i++ ) newItems[ i ] = XmStringCopy( oldItems[ i ] );
	for( j=0; j < numAdds; j++, i++ ) newItems[ i ] = XmStringCopy( addItems[ j ] );

	qsort( (void *)newItems, (size_t)numOld + numAdds, sizeof(XmString), XnListAddSortedXmStringItemXmStringCMP );

	XtSetArg( al[0],		XmNitems,		newItems );
	XtSetArg( al[1],		XmNitemCount,		numOld + numAdds );
	XtSetValues( list, al, 2 );
}

/*******************************************************/

/*
 *  Proc:	XnListGetStringItems
 *  Purpose:	Retrieves the items in a list widget.
 *  Params:	list - The list widget.
 *		numItems - <Return> Number of items in the list.
 *  Returns:	A list of items in the list or NULL if there are none.
 */
StringTable XnListGetStringItems( Widget list, int *numItems )
{
	int		i;
	int		itemCount	=0;
	Arg		al[2];
	StringTable	itemList	=NULL;
	XmStringTable	items		=NULL;

	if( !list || !numItems ) return( NULL );

	(*numItems) = 0;

	XtSetArg( al[0],		XmNitems,		&items );
	XtSetArg( al[1],		XmNitemCount,		&itemCount );
	XtGetValues( list, al, 2 );

	if( !items || !itemCount ) return( NULL );

	itemList = XnMallocStringTable( itemCount );
	for( i=0; i < itemCount; i++ )
		itemList[i] = XnGetXmString( items[i] );

	(*numItems) = itemCount;

	return( itemList );
}

/*******************************************************/

/*
 *  Proc:	XnListSetXmStringItems
 *  Purpose:	Set the list items in a list widget.
 *  Params:	list - The list widget.
 *		items - The new items for the list.
 *		itemCount - The number of items for the list.
 *  Returns:	None
 */
void XnListSetXmStringItems( Widget list, XmStringTable items, int itemCount )
{
	Arg		al[2];

	if( list )
	{
		XtSetArg( al[0],		XmNitems,		items );
		XtSetArg( al[1],		XmNitemCount,		itemCount );
		XtSetValues( list, al, 2 );
	}
}

/*******************************************************/

/*
 *  Proc:	XnListItemCount
 *  Purpose:	Determines the number of items in a list.
 *  Params:	list - The list widget.
 *  Returns:	The number of items in the list.
 */
int XnListItemCount( Widget list )
{
	int		numItems	=0;
	Arg		al[1];

	if( list )
	{
		XtSetArg( al[0],		XmNitemCount,		&numItems );
		XtGetValues( list, al, 1 );
	}

	return( numItems );
}

/*******************************************************/

/*
 *  Proc:	XnScaleSetValues
 *  Purpose:	Set the current value, maximum, and minimum values
 *		for a scale widget.  If any of the arguments are 
 *		negative, their current value is unchanged.
 *  Params:	scale - The scale widget.
 *		value - The new value for the scale.
 *		min - The new minimum value for the scale.
 *		max - The new maximum value for the scale.
 *  Returns:	None
 */
void XnScaleSetValues( Widget scale, int value, int min, int max )
{
	int		ac;
	Arg		al[3];

	ac = 0;

	if( max >= 0 )
	{ 
		XtSetArg( al[ ac ],	XmNmaximum,	max );     
		ac++; 
	}

	if( min >= 0 )
	{ 
		XtSetArg( al[ ac ],	XmNminimum,	min );     
		ac++; 
	}

	if( value >= 0 )
	{ 
		XtSetArg( al[ ac ],	XmNvalue,	value );   
		ac++; 
	}

	XtSetValues( scale, al, ac );
}

/*******************************************************/

/*
 *  Proc:	XnGetUserDataPointer
 *  Purpose:	Get the pointer value associated with a widget.
 *  Params:	widget - The widget.
 *  Returns:	A pointer to the data.
 */
XtPointer XnGetUserDataPointer( Widget widget )
{
	Arg		al[1];
	XtPointer	data	=NULL;

	XtSetArg( al[0],	XmNuserData,	&data );
	XtGetValues( widget, al, 1 );

	return( data );
}

/*******************************************************/

/*
 *  Proc:	XnGetUserDataInt
 *  Purpose:	Get the integer value associated with a widget.
 *  Params:	widget - The widget.
 *  Returns:	The integer value.
 */
int XnGetUserDataInt( Widget widget )
{
	return( *(int *)XnGetUserDataPointer(widget) );
}

/*******************************************************/

/*
 *  Proc:	XnGetUserDataFloat
 *  Purpose:	Get the floating point value associated with a widget.
 *  Params:	widget - The widget.
 *  Returns:	The floating point value.
 */
float XnGetUserDataFloat( Widget widget )
{
	return( *(float *)XnGetUserDataPointer(widget) );
}

/*******************************************************/

/*
 *  Proc:	XnSetUserDataPointer
 *  Purpose:	Sets the pointer value associated with a widget.
 *  Params:	widget - The widget.
 *		data - The data.
 *  Returns:	None
 */
void XnSetUserDataPointer( Widget widget, XtPointer data )
{
	Arg		al[1];

	XtSetArg( al[0],	XmNuserData,	data );
	XtSetValues( widget, al, 1 );
}

/*******************************************************/

/*
 *  Proc:	XnSetUserDataInt
 *  Purpose:	Set the integer value associated with a widget.
 *		The userData storage must already exist.
 *  Params:	widget - The widget.
 *		data - The new integer value.
 *  Returns:	None
 */
void XnSetUserDataInt( Widget widget, int data )
{
	int	*ptr	=NULL;

	ptr = (int *)XnGetUserDataPointer( widget );
	(*ptr) = (int)data;
}

/*******************************************************/

/*
 *  Proc:	XnSetUserDataFloat
 *  Purpose:	Sets the floating point value associated with a widget.
 *		The userData storage must already exist.
 *  Params:	widget - The widget.
 *		data - The new floating point value.
 *  Returns:	None
 */
void XnSetUserDataFloat( Widget widget, float data )
{
	float	*ptr	=NULL;

	ptr = (float *)XnGetUserDataPointer( widget );
	(*ptr) = (float)data;
}

/*******************************************************/

/*
 *  Proc:	XnStrerror
 *  Purpose:	Return a string representing an errno value.
 *  Params:	error - The errno value.
 *  Returns:	An error message.  Some systems may not have
 *		this capability and a generic message is returned.
 */
String XnStrerror( int error )
{
#ifdef sun
	/* return( (String)sys_errlist[ error ] ); */
	return( "(Error Type Not Available)" );
#else
	return( (String)strerror(error) );
#endif
}

/*******************************************************/

/*
 *  Proc:	XnExecCommand
 *  Purpose:	Executes a command and returns its output in strings.
 *		See NXUtils.h for a more detailed description of the function.
 *  Params:	command - The command to run.
 *		argv - The arguments for the command (like exec(3)).
 *		stdinfile - A file for standard input.
 *		stdoutfile - A file for standard output.
 *		stderrfile - A file for stderr output.
 *		stdinput - A string for standard input (if !stdinfile).
 *		stdoutput - <Return> A string that will contain stdout.
 *		stderrput - <Return> A string that will contain stderr.
 *		exitCode - <Return> The exit code from the process.
 *		background - True if the command is not to be waited on.
 *		removeFiles - True if the files above are to be removed after use.
 *  Returns:	True if the command was executed successfully
 *		(includes the reading of files into strings).
 */
Boolean XnExecCommand( String command, String argv[],
			String stdinfile, String stdoutfile, String stderrfile,
			String stdinput, String *stdoutput, String *stderrput,
			int *exitCode, Boolean background, Boolean removeFiles )
{
	int	i;
	char	infile[ 256 ];
	char	outfile[ 256 ];
	char	errfile[ 256 ];
	String	commandLine	=NULL;
	Boolean	error		=False;
	Boolean	tmpinfile	=False;
	Boolean	tmpoutfile	=False;
	Boolean	tmperrfile	=False;

	(*exitCode) = 0;

	if( !command || !argv || !exitCode ) return( False );

	error = False;
	tmpinfile = False;
	tmpoutfile = False;
	tmperrfile = False;

	if( stdoutput ) (*stdoutput) = NULL;
	if( stderrput ) (*stderrput) = NULL;

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/*
	 * If they didn't specify an input file, see if they
	 * want to specify an input string.  If they do then write
	 * the text to a file that can be redirected. Otherwise use
	 * dev/null.
	 */
	if( !stdinfile )
	{
		if( stdinput )
		{
			tmpnam( infile );
			XnWriteFile( infile, stdinput, "w", &i );
			tmpinfile = True;
		}
		else strcpy( infile, "/dev/null" );
	}
	else strcpy( infile, stdinfile );

	/*
	 * If they didn't specify an output file but they still
	 * want the output then we have to create a temporary file
	 * to use for stdout.
	 */
	if( !stdoutfile )
	{
		if( stdoutput )
		{
			tmpnam( outfile );
			tmpoutfile = True;
		}
		else strcpy( outfile, "/dev/null" );
	}
	else strcpy( outfile, stdoutfile );

	/*
	 * If they didn't specify an error file but still want the
	 * error output then we have to create a temporary file to
	 * use for stderr.
	 */
	if( !stderrfile )
	{
		if( stderrput )
		{
			tmpnam( errfile );
			tmperrfile = True;
		}
		else strcpy( errfile, "/dev/null" );
	}
	else strcpy( errfile, stderrfile );

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/*
	 * Build the command string then execute it.
	 */
	commandLine = XnDupString( command );

	if( argv[0] && argv[1] )
	{
		for( i=1; argv[i]; i++ )
		{
			commandLine = XnExtendString( commandLine, " " );
			commandLine = XnExtendString( commandLine, argv[i] );
		}
	}

	/*
	 * If the empty string is used as a filename then
	 * don't redirect that source.
 	 */
	if( infile[0] )
	{
		commandLine = XnExtendString( commandLine, " 0<" );
		commandLine = XnExtendString( commandLine, infile );
	}

	if( outfile[0] )
	{
		commandLine = XnExtendString( commandLine, " 1>" );
		commandLine = XnExtendString( commandLine, outfile );
	}

	if( errfile[0] )
	{
		commandLine = XnExtendString( commandLine, " 2>" );
		commandLine = XnExtendString( commandLine, errfile );
	}

	commandLine = XnExtendString( commandLine, " " );

	if( background ) commandLine = XnExtendString( commandLine, "&" );

	(*exitCode) = system( commandLine );

	/* Return immediately if running in background. */
	if( background )
	{
		/*
		 * If they wanted to remove the files - do so.
		 */
		if( tmpinfile ) XnDeleteFile( infile );
		if( tmpoutfile ) XnDeleteFile( outfile );
		if( tmperrfile ) XnDeleteFile( errfile );

		if( removeFiles )
		{
			if( stdinfile ) XnDeleteFile( stdinfile );
			if( stdoutfile ) XnDeleteFile( stdoutfile );
			if( stderrfile ) XnDeleteFile( stderrfile );
		}

		if( (*exitCode) == 127 ) return( False );

		return( !error );
	}

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/*
	 * Read the text in the stdout file into a string.
	 */
	i = 0;
	if( stdoutput && outfile[0] ) (*stdoutput) = XnReadFile( outfile, NULL, &i );
	if( i ) error = True;

	/*
	 * Read the text in the stderr file into a string.
	 */
	i = 0;
	if( stderrput && errfile[0] ) (*stderrput) = XnReadFile( errfile, NULL, &i );
	if( i ) error = True;

	/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	/*
	 * If they wanted to remove the files - do so.
	 */
	if( tmpinfile ) XnDeleteFile( infile );
	if( tmpoutfile ) XnDeleteFile( outfile );
	if( tmperrfile ) XnDeleteFile( errfile );

	if( removeFiles )
	{
		if( stdinfile ) XnDeleteFile( stdinfile );
		if( stdoutfile ) XnDeleteFile( stdoutfile );
		if( stderrfile ) XnDeleteFile( stderrfile );
	}

	/*
	 * If we couldn't system the command return False.
	 */
	if( (*exitCode) == 127 ) return( False );

	return( !error );
}

/*******************************************************/

/*
 *  Proc:	XnReadFile
 *  Purpose:	Read a text file into a string.
 *  Params:	filename - The name of the file.
 *		defaultValue - A default value if the file can't be read.
 *		retVal - <Return> A return exit status.
 *  Returns:	The contents of the file.  If the file is empty
 *		or there are problems reading the file, it returns
 *		a default value (if nonNULL).  retVal contains
 *		any exit codes and is 0 if there are no errors.
 */
String XnReadFile( String filename, String defaultValue, int *retVal )
{
	int	size	=0;
	String	text	=NULL;
	FILE	*file	=NULL;

	if( retVal ) (*retVal) = -1;

	if( !filename ) return( NULL );

	if( retVal ) (*retVal) = 0;

	if( (size=XnFileSize(filename)) > 0 )
	{
		if( (file = fopen(filename, "r")) != NULL )
		{
			text = XnMallocString( size + 1 );

			if( fread( (void *)text, (size_t)size, sizeof(char), file ) )
			{
				text[ size - 1 ] = '\0';
			}
			else
			{
				if( retVal ) (*retVal) = errno;
				XnFreeString( text );
				text = NULL;
			}

			fclose( file );

			return( text );
		}
		else if( retVal ) (*retVal) = errno;
	}

	if( defaultValue ) return( XnDupString(defaultValue) );

	return( NULL );
}

/*******************************************************/

/*
 *  Proc:	XnWriteFile
 *  Purpose:	Write text to a file.
 *  Params:	filename - The name of the file.
 *		text - The text to be written.
 *		writemode - The mode to open the file for writing.
 *		retVal - <Return> A return exit status.
 *  Returns:	The number of characters written.
 */
int XnWriteFile( String filename, String text, String writemode, int *retVal )
{
	int	written	=0;
	FILE	*file	=NULL;

	written = 0;

	if( retVal ) (*retVal) = -1;

	if( !filename || !text ) return( 0 );

	if( retVal ) (*retVal) = 0;

	if( (file = fopen(filename, writemode)) != NULL )
	{
		written = fwrite( (void *)text, strlen(text), sizeof(char), file );

		fclose( file );
	}
	else if( retVal ) (*retVal) = errno;

	return( written );
}


